import numpy as np
import pytest

from el0ps.compilation import CompilableClass, compiled_clone
from el0ps.penalty import (
    BasePenalty,
    SymmetricPenalty,
    Bigm,
    BigmL1norm,
    BigmL2norm,
    BigmPositiveL1norm,
    BigmPositiveL2norm,
    Bounds,
    L1norm,
    L2norm,
    L1L2norm,
    PositiveL1norm,
    PositiveL2norm,
    compute_param_slope_pos,
    compute_param_slope_neg,
)

n = 100
x = np.random.randn(n)
u = np.random.randn(n)
lmbd = np.random.rand()
bigm = np.linalg.norm(x, np.inf)
alpha = np.random.rand()
beta = np.random.rand()
penalties = [
    pytest.param(penalty, id=f"{penalty.__class__.__name__}")
    for penalty in [
        Bigm(bigm),
        BigmL1norm(bigm, alpha),
        BigmL2norm(bigm, alpha),
        BigmPositiveL1norm(bigm, alpha),
        BigmPositiveL2norm(bigm, beta),
        Bounds(-2 * np.ones(n), 2 * np.ones(n)),
        L1norm(alpha),
        L2norm(alpha),
        L1L2norm(alpha, beta),
        PositiveL1norm(alpha),
        PositiveL2norm(beta),
    ]
]


@pytest.mark.parametrize("penalty", penalties)
def test_penalty(penalty: BasePenalty):

    if isinstance(penalty, CompilableClass):
        penalty = compiled_clone(penalty)

    assert isinstance(penalty.__str__(), str)

    for i, (xi, ui) in enumerate(zip(x, u)):

        assert penalty.value(i, 0.0) == 0.0
        assert penalty.value(i, xi) >= 0.0
        assert penalty.conjugate(i, ui) >= 0.0
        assert (
            penalty.value(i, xi) + penalty.conjugate(i, ui) >= xi * ui - 1e-10
        )

        tol = 1e-8
        slope_pos = penalty.param_slope_pos(i, lmbd)
        slope_neg = penalty.param_slope_neg(i, lmbd)
        slope_pos_approx = compute_param_slope_pos(penalty, i, lmbd, tol=tol)
        slope_neg_approx = compute_param_slope_neg(penalty, i, lmbd, tol=tol)
        assert slope_pos >= 0.0
        assert slope_neg <= 0.0
        assert slope_pos == pytest.approx(slope_pos_approx, abs=tol)
        assert slope_neg == pytest.approx(slope_neg_approx, abs=tol)

        limit_pos = penalty.param_limit_pos(i, lmbd)
        limit_neg = penalty.param_limit_neg(i, lmbd)
        if limit_pos == 0.0:
            assert slope_pos == np.inf
        elif limit_pos < np.inf:
            assert penalty.conjugate(i, slope_pos) == pytest.approx(lmbd)
            assert (
                penalty.value(i, limit_pos) + penalty.conjugate(i, slope_pos)
                >= limit_pos * slope_pos - 1e-10
            )
        else:
            assert penalty.conjugate(i, slope_pos) < lmbd
        if limit_neg == 0.0:
            assert slope_neg == -np.inf
        elif limit_neg > -np.inf:
            assert penalty.conjugate(i, slope_neg) == pytest.approx(lmbd)
            assert (
                penalty.value(i, limit_neg) + penalty.conjugate(i, slope_neg)
                >= limit_neg * slope_neg - 1e-10
            )
        else:
            assert penalty.conjugate(i, slope_neg) < lmbd

        if isinstance(penalty, SymmetricPenalty):
            assert slope_pos == pytest.approx(-slope_neg)
            assert limit_pos == pytest.approx(-limit_neg)

        eta = np.random.rand()
        pi = penalty.prox(i, xi, eta)
        v1 = 0.5 * (pi - xi) ** 2 + eta * penalty.value(i, pi)
        v2 = eta * penalty.value(i, xi)
        assert v1 <= v2



import numpy as np
from el0ps.penalty import (
    Bigm,
    BigmL1norm,
    BigmL2norm,
    L1L2norm,
    L1norm,
    L2norm,
)


def select_bigml1l2_penalty(
    alpha: float = 0.0, beta: float = 0.0, M: float = np.inf
):
    if alpha == 0.0 and beta == 0.0 and M != np.inf:
        penalty = Bigm(M)
    elif alpha != 0.0 and beta == 0.0 and M == np.inf:
        penalty = L1norm(alpha)
    elif alpha != 0.0 and beta == 0.0 and M != np.inf:
        penalty = BigmL1norm(M, alpha)
    elif alpha == 0.0 and beta != 0.0 and M == np.inf:
        penalty = L2norm(beta)
    elif alpha == 0.0 and beta != 0.0 and M != np.inf:
        penalty = BigmL2norm(M, beta)
    elif alpha != 0.0 and beta != 0.0 and M == np.inf:
        penalty = L1L2norm(alpha, beta)
    else:
        raise ValueError(
            "Setting `alpha=0`, `beta=0` and `M=np.inf` simultaneously is not "
            "allowed."
        )
    return penalty



"""Base classes for penalty functions and related utilities."""

import numpy as np
import pyomo.kernel as pmo
from abc import abstractmethod
from numpy.typing import NDArray


class BasePenalty:
    r"""Base class for penalty defined as separable functions.

    This class represent separable mathematical functions expressed as

    .. math::
        \begin{align*}
            h : \mathbb{R}^n &\rightarrow \mathbb{R} \cup \{+\infty\} \\
            \mathbf{x} &\mapsto h(\mathbf{x}) = \textstyle\sum_{i=1}^n h_i(x_i)
        \end{align*}


    where each splitting term :math:`h_i` is proper, lower-semicontinuous,
    convex, coercive, non-negative, and minimized at :math:`x_i = 0`.
    """

    @abstractmethod
    def value(self, i: int, x: float) -> float:
        """Value of the i-th splitting term of the penalty function at ``x``.

        Parameters
        ----------
        i : int
            Index of the splitting term.
        x : float
            Value at which the splitting term is evaluated.

        Returns
        -------
        value : float
            The value of the i-th splitting term of the function at ``x``.
        """
        ...

    @abstractmethod
    def conjugate(self, i: int, x: float) -> float:
        """Value of the convex conjugate of the i-th splitting term of the
        penalty function at ``x``.

        Parameters
        ----------
        i : int
            Index of the splitting term.
        x : float
            Value at which the convex conjugate is evaluated.

        Returns
        -------
        value : float
            The value of the convex conjugate.
        """
        ...

    @abstractmethod
    def prox(self, i: int, x: float, eta: float) -> float:
        """Proximity operator of the i-th splitting term of the penalty
        function weighted by ``eta`` at ``x``.

        Parameters
        ----------
        i : int
            Index of the splitting term.
        x : float
            Value at which the proximal operator is evaluated.
        eta : float, positive
            Multiplicative weight.

        Returns
        -------
        value : float
            The value of the proximity operator.
        """
        ...

    @abstractmethod
    def subdiff(self, i: int, x: float) -> NDArray:
        """Subdifferential of the i-th splitting term of the penalty function
        at ``x``, returned as an interval.

        Parameters
        ----------
        i : int
            Index of the splitting term.
        x : float
            Value at which the subdifferential is evaluated.

        Returns
        -------
        value : NDArray
            1D-array of size 2 containing the lower and upper bounds of the
            interval corresponding to the subdifferential.
        """
        ...

    @abstractmethod
    def conjugate_subdiff(self, i: int, x: float) -> NDArray:
        """Subdifferential of the conjugate of the i-th splitting term of the
        penalty function at ``x``, returned as an interval.

        Parameters
        ----------
        i : int
            Index of the splitting term.
        x : float
            Value at which the subdifferential of the conjugate is evaluated.

        Returns
        -------
        value : NDArray
            1D-array of size 2 containing the lower and upper bounds of the
            interval corresponding to the conjugate subdifferential.
        """
        ...

    def param_slope_pos(self, i: int, lmbd: float) -> float:
        """Supremum of the set ``{x in R | self.conjugate(i, x) <= lmbd}``.

        Parameters
        ----------
        i : int
            Index of the splitting term in the set definition.
        lmbd : float
            Threshold value in the set definition.

        Returns
        -------
        value : float
            The supremum of the set.
        """
        return compute_param_slope_pos(self, i, lmbd)

    def param_slope_neg(self, i: int, lmbd: float) -> float:
        """Infimum of the set ``{x in R | self.conjugate(i, x) <= lmbd}``.

        Parameters
        ----------
        i : int
            Index of the splitting term in the set definition.
        lmbd : float
            Threshold value in the set definition.

        Returns
        -------
        value : float
            The infimum of the set.
        """
        return compute_param_slope_neg(self, i, lmbd)

    def param_limit_pos(self, i: int, lmbd: float) -> float:
        """Supremum of the set ``self.conjugate_subdiff(i, tau)`` where
        ``tau = self.param_slope_pos(i, lmbd)``.

        Parameters
        ----------
        i : int
            Index of the splitting term in the definition of ``tau``.
        lmbd : float
            Threshold value in the definition of ``tau``.

        Returns
        -------
        value : float
            The supremum of the set.
        """
        s = self.conjugate_subdiff(i, self.param_slope_pos(i, lmbd))
        return np.inf if np.all(np.isnan(s)) else s[1]

    def param_limit_neg(self, i: int, lmbd: float) -> float:
        """Infimum of the set ``self.conjugate_subdiff(i, tau)`` where
        ``tau = self.param_slope_neg(i, lmbd)``.

        Parameters
        ----------
        i : int
            Index of the splitting term in the definition of ``tau``.
        lmbd : float
            Threshold value in the definition of ``tau``.

        Returns
        -------
        value : float
            The infimum of the set.
        """
        s = self.conjugate_subdiff(i, self.param_slope_neg(i, lmbd))
        return -np.inf if np.all(np.isnan(s)) else s[0]

    def param_bndry_pos(self, i: int, lmbd: float) -> float:
        """Supremum of the set ``self.subdiff(i, tau)`` where
        ``tau = self.param_limit_pos(i, lmbd)``.

        Parameters
        ----------
        i : int
            Index of the splitting term in the definition of ``tau``.
        lmbd : float
            Threshold value in the definition of ``tau``.

        Returns
        -------
        value : float
            The supremum of the set.
        """
        tau = self.param_limit_pos(i, lmbd)
        return np.inf if tau == np.inf else self.subdiff(i, tau)[1]

    def param_bndry_neg(self, i: int, lmbd: float) -> float:
        """Infimum of the set ``self.subdiff(i, tau)`` where
        ``tau = self.param_limit_neg(i, lmbd)``.

        Parameters
        ----------
        i : int
            Index of the splitting term in the definition of ``tau``.
        lmbd : float
            Threshold value in the definition of ``tau``.

        Returns
        -------
        value : float
            The supremum of the set.
        """
        tau = self.param_limit_neg(i, lmbd)
        return -np.inf if tau == -np.inf else self.subdiff(i, tau)[0]


class SymmetricPenalty(BasePenalty):
    """Base class for symmetric :class:`BasePenalty` functions."""

    @abstractmethod
    def param_slope(self, i: int, lmbd: float) -> float:
        """Supremum of the set ``{x in R | self.conjugate(i, x) <= lmbd}``.

        Parameters
        ----------
        i : int
            Index of the splitting term in the set definition.
        lmbd : float
            Threshold value in the set definition.

        Returns
        -------
        value : float
            The supremum of the set.
        """
        ...

    def param_slope_pos(self, i: int, lmbd: float) -> float:
        return self.param_slope(i, lmbd)

    def param_slope_neg(self, i: int, lmbd: float) -> float:
        return -self.param_slope(i, lmbd)

    def param_limit(self, i: int, lmbd: float) -> float:
        s = self.conjugate_subdiff(i, self.param_slope(i, lmbd))
        return np.inf if np.all(np.isnan(s)) else s[1]

    def param_limit_pos(self, i, lmbd):
        return self.param_limit(i, lmbd)

    def param_limit_neg(self, i, lmbd):
        return -self.param_limit(i, lmbd)

    def param_bndry(self, i: int, lmbd: float) -> float:
        tau = self.param_limit(i, lmbd)
        return np.inf if tau < np.inf else self.subdiff(i, tau)[1]

    def param_bndry_pos(self, i: int, lmbd: float) -> float:
        return self.param_bndry(i, lmbd)

    def param_bndry_neg(self, i: int, lmbd: float) -> float:
        return -self.param_bndry(i, lmbd)


class MipPenalty:
    """Base class for penalty functions that can be modeled into
    `pyomo <https://pyomo.readthedocs.io/en/stable/>`_."""

    @abstractmethod
    def bind_model(self, model: pmo.block) -> None:
        """Impose an constraint associated with the penalty function in a
        `pyomo <https://pyomo.readthedocs.io/en/stable/>`_ model.

        Given a pyomo.kernel.block ``model`` object containing a real scalar
        variable ``model.h`` and a real vector variable ``model.x`` of size
        ``model.N``, this function is intended to impose the relations

        ``model.h >= self.value(model.x)``

        and

        ``model.z[i] = 0 ==> model.x[i] = 0 for all i in model.N``

        using ``pyomo`` expressions.

        Parameters
        ----------
        model : pmo.block
            The pyomo kernel model.
        lmbd : float
            The scalar parameter involved in the expressions to be modeled.
        """
        ...


def compute_param_slope_pos(
    penalty: BasePenalty,
    i: int,
    lmbd: float,
    tol: float = 1e-8,
    maxit: int = 100,
) -> float:
    """Utility to approximate the value of the function
    ``penalty.param_slope_pos`` for a given :class:`BasePenalty` instance using
    a bisection method.

    Parameters
    ----------
    penalty : BasePenalty
        The penalty function.
    i : int
        Parameter involved in the ``penalty.param_slope_pos`` function.
    lmbd : float
        Parameter involved in the ``penalty.param_slope_pos`` function.
    tol : float = 1e-4
        Bisection approximation tolerance.
    maxit : int = 100
        Maximum number of bisection iterations.
    """
    a = 0.0
    b = 1.0
    while penalty.conjugate(i, b) < lmbd:
        b *= 2.0
        if b > 1e12:
            return np.inf
    for _ in range(maxit):
        c = (a + b) / 2.0
        fa = penalty.conjugate(i, a) - lmbd
        fc = penalty.conjugate(i, c) - lmbd
        if (-tol <= fc <= tol) or (b - a < 0.5 * tol):
            return c
        elif fc * fa >= 0.0:
            a = c
        else:
            b = c
    return c


def compute_param_slope_neg(
    penalty: BasePenalty,
    i: int,
    lmbd: float,
    tol: float = 1e-8,
    maxit: int = 100,
) -> float:
    """Utility to approximate the value of the function
    ``penalty.param_slope_neg`` for a given :class:`BasePenalty` instance using
    a bisection method.

    Parameters
    ----------
    penalty : BasePenalty
        The penalty function.
    i : int
        Parameter involved in the ``penalty.param_slope_neg`` function.
    lmbd : float
        Parameter involved in the ``penalty.param_slope_neg`` function.
    tol : float = 1e-4
        Bisection approximation tolerance.
    maxit : int = 100
        Maximum number of bisection iterations.
    """
    a = -1.0
    b = 0.0
    while penalty.conjugate(i, a) < lmbd:
        a *= 2.0
        if a < -1e12:
            return -np.inf
    for _ in range(maxit):
        c = (a + b) / 2.0
        fa = penalty.conjugate(i, a) - lmbd
        fc = penalty.conjugate(i, c) - lmbd
        if (-tol <= fc <= tol) or (b - a < 0.5 * tol):
            return c
        elif fc * fa >= 0.0:
            a = c
        else:
            b = c
    return c



"""Scikit-learn-compatible estimators stemming from L0-regularized problems."""

from .base import L0Estimator
from .classifier import (
    L0Classifier,
    L0L1Classifier,
    L0L2Classifier,
    L0L1L2Classifier,
)
from .regressor import (
    L0Regressor,
    L0L1Regressor,
    L0L2Regressor,
    L0L1L2Regressor,
)
from .svc import (
    L0SVC,
    L0L1SVC,
    L0L2SVC,
    L0L1L2SVC,
)

__all__ = [
    "L0Estimator",
    "L0Classifier",
    "L0L1Classifier",
    "L0L2Classifier",
    "L0L1L2Classifier",
    "L0Regressor",
    "L0L1Regressor",
    "L0L2Regressor",
    "L0L1L2Regressor",
    "L0SVC",
    "L0L1SVC",
    "L0L2SVC",
    "L0L1L2SVC",
]



import numpy as np

from el0ps.datafit import Leastsquares
from el0ps.penalty import Bigm
from el0ps.solver import BnbSolver, Status
from el0ps.path import Path
from .utils import make_regression


def test_path():
    A, y, x_true = make_regression(3, 150, 200)
    M = 1.5 * np.max(np.abs(x_true))

    datafit = Leastsquares(y)
    penalty = Bigm(M)
    solver = BnbSolver(verbose=False)

    path = Path(
        lmbd_max=1e-0, lmbd_min=1e-1, lmbd_num=10, lmbd_normalized=True
    )
    fit = path.fit(solver, datafit, penalty, A)
    assert np.all(result.status == Status.OPTIMAL for _, result in fit.items())



import numpy as np
from numpy.typing import NDArray
from sklearn.multiclass import check_classification_targets
from sklearn.base import BaseEstimator
from sklearn.linear_model._base import LinearModel, LinearClassifierMixin
from sklearn.preprocessing import LabelEncoder
from sklearn.utils.validation import validate_data

from el0ps.solver import BaseSolver, BnbSolver
from el0ps.datafit import BaseDatafit
from el0ps.penalty import BasePenalty


class L0Estimator(LinearModel, BaseEstimator):
    r"""Scikit-learn-compatible `linear model <https://scikit-learn.org/stable/api/sklearn.linear_model.html>`_
    estimators based on L0-regularized problems.

    The estimator corresponds to a solution of the problem

    .. math::

        \textstyle\min_{\mathbf{x} \in \mathbb{R}^{n}} f(\mathbf{Ax}) + \lambda\|\mathbf{x}\|_0 + h(\mathbf{x})

    where :math:`f` is a datafit function,
    :math:`\mathbf{A} \in \mathbb{R}^{m \times n}` is a matrix,
    :math:`\lambda > 0` is a parameter, the L0-norm :math:`\|\cdot\|_0` counts
    the number of non-zero entries in its input, and :math:`h` is a penalty
    function.

    Parameters
    ----------
    datafit: BaseDatafit
        Datafit function.
    penalty: BasePenalty
        Penalty function.
    lmbd: float
        L0-norm weight.
    fit_intercept: bool, default=False
        Whether to fit an intercept term.
    solver: BaseSolver, default=BnbSolver()
        Solver for the estimator associated problem.
    """  # noqa: E501

    def __init__(
        self,
        datafit: BaseDatafit,
        penalty: BasePenalty,
        lmbd: float,
        fit_intercept: bool = False,
        solver: BaseSolver = BnbSolver(),
    ) -> None:

        if fit_intercept:
            raise NotImplementedError("Fit intercept not implemented yet")

        if not hasattr(datafit, "y"):
            raise ValueError(
                "The datafit object must have an attribute `y` to be used in "
                "an `L0Estimator`."
            )

        self.datafit = datafit
        self.penalty = penalty
        self.lmbd = lmbd
        self.fit_intercept = fit_intercept
        self.solver = solver

        self.is_fitted_ = False
        self.coef_ = None
        self.intercept_ = None
        self.n_iter_ = None

    def fit(self, X: NDArray, y: NDArray):

        # Sanity checks
        if isinstance(self, LinearClassifierMixin):
            check_classification_targets(y)
            enc = LabelEncoder()
            y = enc.fit_transform(y)
            y = 2.0 * y - 1.0
            if len(enc.classes_) > 2:
                raise ValueError("Only binary classification is supported.")
            self.classes_ = np.unique(y)
        else:
            self.classes_ = None
        check_X_params = dict(dtype=np.float64, order="F")
        check_y_params = dict(dtype=np.float64, order="F", ensure_2d=False)
        X, y = validate_data(
            self, X, y, validate_separately=(check_X_params, check_y_params)
        )
        assert y.ndim == 1

        # Initialize estimator.coef_
        if self.coef_ is None:
            self.coef_ = np.zeros(X.shape[1])

        # Initialize estimator.intercept_
        if self.intercept_ is None:
            self.intercept_ = 0.0

        # Update datafit target vector
        y_old = np.copy(self.datafit.y)
        self.datafit.y = y

        # Solve the estimator optimization problem
        result = self.solver.solve(
            self.datafit, self.penalty, X, self.lmbd, x_init=self.coef_
        )

        # Reset the datafit target vector
        self.datafit.y = y_old

        # Recover the results
        self.is_fitted_ = True
        self.fit_result_ = result
        self.coef_ = np.copy(result.x)
        self.n_iter_ = result.iter_count
        # TODO: Recover intercept when supported
        self.intercept_ = 0.0

        return self

    def predict(self, X: NDArray) -> NDArray:
        if not self.is_fitted_:
            raise ValueError("The estimator is not fitted yet.")
        decision = X @ self.coef_ + self.intercept_
        if isinstance(self, LinearClassifierMixin):
            binary_preds = (decision > 0).astype(int)
            decision = self.classes_[binary_preds]
        return decision



"""Base classes for datafit functions and related utilities."""

import pyomo.kernel as pmo
from abc import abstractmethod
from numpy.typing import NDArray


class BaseDatafit:
    r"""Base class for datafit functions.

    This class represent mathematical functions expressed as

    .. math::
        \begin{align*}
            f : \mathbb{R}^m &\rightarrow \mathbb{R} \cup \{+\infty\} \\
            \mathbf{w} &\mapsto f(\mathbf{w})
        \end{align*}

    that are proper, lower-semicontinuous, convex, and differentiable with a
    Lipschitz-continuous gradient."""

    @abstractmethod
    def value(self, w: NDArray) -> float:
        """Value of the function at ``w``.

        Parameters
        ----------
        w : NDArray
            Vector at which the function is evaluated.

        Returns
        -------
        value : float
            The function value at ``w``.
        """
        ...

    @abstractmethod
    def conjugate(self, w: NDArray) -> float:
        """Value of the convex conjugate of the function at ``w``.

        Parameters
        ----------
        w : NDArray
            Vector at which the conjugate is evaluated.

        Returns
        -------
        value : float
            The conjugate value at ``w``.
        """
        ...

    @abstractmethod
    def gradient(self, w: NDArray) -> NDArray:
        """Value of gradient at ``w``.

        Parameters
        ----------
        w : NDArray
            Vector at which the gradient is evaluated.

        Returns
        -------
        value : NDArray
            The gradient at ``w``.
        """
        ...

    @abstractmethod
    def gradient_lipschitz_constant(self) -> float:
        """Lipschitz constant of the gradient.

        Returns
        -------
        value : float
            The Lipschitz constant of the gradient.
        """
        ...


class MipDatafit:
    """Base class for datafit functions that can be modeled into
    `pyomo <https://pyomo.readthedocs.io/en/stable/>`_."""

    @abstractmethod
    def bind_model(self, model: pmo.block) -> None:
        """Impose a constraint associated with the datafit function in a
        `pyomo <https://pyomo.readthedocs.io/en/stable/>`_ model.

        Given a pyomo.kernel.block ``model`` object containing a real scalar
        variable ``model.f`` and a real vector variable ``model.w`` of size
        ``model.M``, this function is intended to imposer the relation

        ``model.f >= self.value(model.w)``

        using ``pyomo`` expressions.

        Parameters
        ----------
        model : pmo.block
            The pyomo kernel model.
        """
        ...



import pytest
import numpy as np
from sklearn.utils.estimator_checks import check_estimator
from el0ps.datafit import Leastsquares, Logistic, Squaredhinge
from el0ps.estimator import (
    L0L1L2Regressor,
    L0L1L2Classifier,
    L0L1L2SVC,
)
from el0ps.penalty import L1L2norm
from el0ps.solver import Status
from el0ps.utils import compute_lmbd_max
from .utils import make_classification, make_regression, make_svc

k, m, n = 3, 15, 20
lmbd_factor = 0.1
alpha = 0.2
beta = 0.3

A_cls, y_cls, coef_true_cls = make_classification(k, m, n)
lmbd_cls = lmbd_factor * compute_lmbd_max(
    Logistic(y_cls), L1L2norm(alpha, beta), A_cls
)

A_reg, y_reg, coef_true_reg = make_regression(k, m, n)
lmbd_reg = lmbd_factor * compute_lmbd_max(
    Leastsquares(y_reg), L1L2norm(alpha, beta), A_reg
)

A_svc, y_svc, coef_true_svc = make_svc(k, m, n)
lmbd_svc = lmbd_factor * compute_lmbd_max(
    Squaredhinge(y_svc), L1L2norm(alpha, beta), A_svc
)

test_data = [
    pytest.param(estimator, A, y, id=f"{estimator.__class__.__name__}")
    for estimator, A, y in [
        (L0L1L2Classifier(lmbd_cls, alpha, beta), A_cls, y_cls),
        (L0L1L2Regressor(lmbd_reg, alpha, beta), A_reg, y_reg),
        (L0L1L2SVC(lmbd_cls, alpha, beta), A_svc, y_svc),
    ]
]


@pytest.mark.parametrize("estimator,A,y", test_data)
def test_estimator(estimator, A, y):
    checks = check_estimator(estimator, generate_only=True)
    assert np.all(checks)
    estimator.fit(A, y)
    y_pred = estimator.predict(A)
    estimator.score(A, y)
    assert estimator.fit_result_.status == Status.OPTIMAL
    assert y_pred.shape == y.shape



import sys
import pybnb as pb
import numpy as np
from copy import deepcopy
from dataclasses import dataclass
from numba import boolean, float64, int64
from numba.experimental.jitclass.base import JitClassType
from numpy.typing import NDArray
from typing import Optional, Union

from el0ps.datafit import BaseDatafit
from el0ps.penalty import BasePenalty
from el0ps.compilation import CompilableClass, compiled_clone
from el0ps.solver import Status, Result


@dataclass
class NodeState:
    category: int
    S0: NDArray
    S1: NDArray
    Sb: NDArray
    x_lower: NDArray
    x_upper: NDArray
    upper_bound: float
    lower_bound: float

    def fix_to(self, idx: int, val: bool):
        self.Sb[idx] = False
        if val:
            self.category = 1
            self.S1[idx] = True
        else:
            self.category = 0
            self.S0[idx] = True
            self.x_lower[idx] = 0.0
            self.x_upper[idx] = 0.0


@dataclass
class TreeState:
    best_upper_bound: float


class BoundSolver(CompilableClass):
    r"""Bounding solver for the :class:`BnbSolver`.

    The solver solves convex optimization problems of the form

    .. math::

        \textstyle\min_{\mathbf{x} \in \mathbb{R}^{n}} f(\mathbf{Ax}) + g(\mathbf{x})

    where :math:`f` is a :class:`el0ps.datafit.BaseDatafit` function,
    :math:`\mathbf{A} \in \mathbb{R}^{m \times n}` is a matrix, :math:`g` is a
    regularization function that depends on the current node treated by the
    :class:`BnbSolver` solver. The problem is solved using a coordinate-descent
    method, potentially combined with an active-set strategy.

    Parameters
    ----------
    iter_limit: int, default=sys.maxsize
        Maximum number of iterations for the inner solver.
    relative_tol: float, default=1e-4
        Relative tolerance on the problem objective value.
    workingset: bool, default=True
        Toggle the use of a working-set method.
    dualpruning: bool, default=True
        If ``True``, the solver checks after each working-set update if the
        dual bound is greater than the best upper bound known during the
        Branch-and-Bound algorithm. If so, the node currently treated is
        pruned. See "Techniques for accelerating Branch-and-Bound algorithms
        dedicated to sparse optimization" by G. Samain et al. for more details.
    screening: bool, default=True
        If ``True``, the solver performs screening tests after each working-set
        update to identify variables that can be fixed to zero to reduce the
        computational load. See "One to beat them all: RYU--a unifying
        framework for the construction of safe balls" by T.L. Tran et al. for
        more details.
    simpruning: bool, default=True
        If ``True``, the solver performs simultaneous pruning tests after each
        working-set update. This allows to identify new branchings that can be
        performed on the current node. This in done in-place during the
        bounding process. See "A New Branch-and-bound pruning framework for
        L0-regularized problems" by T. Guyard et al. for more details.
    """  # noqa: E501

    def __init__(
        self,
        iter_limit: Optional[int] = None,
        relative_tol: float = 1e-4,
        workingset: bool = True,
        dualpruning: bool = True,
        screening: bool = True,
        simpruning: bool = True,
    ) -> None:

        # Solver parameters
        self.iter_limit = iter_limit if iter_limit is not None else sys.maxsize
        self.relative_tol = relative_tol
        self.workingset = workingset
        self.dualpruning = dualpruning
        self.screening = screening
        self.simpruning = simpruning

        # Working values
        self.S0 = np.empty(0, dtype=np.bool_)
        self.S1 = np.empty(0, dtype=np.bool_)
        self.Sb = np.empty(0, dtype=np.bool_)
        self.Sbi = np.empty(0, dtype=np.bool_)
        self.Ws = np.empty(0, dtype=np.bool_)
        self.x = np.empty(0, dtype=np.float64)
        self.w = np.empty(0, dtype=np.float64)
        self.u = np.empty(0, dtype=np.float64)
        self.v = np.empty(0, dtype=np.float64)
        self.pv = np.inf
        self.pv_old = np.inf
        self.dv = np.nan
        self.ub = np.inf
        self.lipschitz = np.nan
        self.A_colnorm = np.empty(0, dtype=np.float64)
        self.stepsize = np.empty(0, dtype=np.float64)
        self.param_slope_pos = np.empty(0, dtype=np.float64)
        self.param_slope_neg = np.empty(0, dtype=np.float64)
        self.param_limit_pos = np.empty(0, dtype=np.float64)
        self.param_limit_neg = np.empty(0, dtype=np.float64)
        self.inner_relative_tol = 0.1 * self.relative_tol
        self.flag_Ws_update = False
        self.upper = False

    def get_spec(self) -> tuple:
        spec = (
            ("iter_limit", int64),
            ("relative_tol", float64),
            ("workingset", boolean),
            ("dualpruning", boolean),
            ("screening", boolean),
            ("simpruning", boolean),
            ("S0", boolean[:]),
            ("S1", boolean[:]),
            ("Sb", boolean[:]),
            ("Sbi", boolean[:]),
            ("Ws", boolean[:]),
            ("x", float64[::1]),
            ("w", float64[::1]),
            ("u", float64[::1]),
            ("v", float64[::1]),
            ("pv", float64),
            ("pv_old", float64),
            ("dv", float64),
            ("ub", float64),
            ("lipschitz", float64),
            ("A_colnorm", float64[:]),
            ("stepsize", float64[:]),
            ("param_slope_pos", float64[:]),
            ("param_slope_neg", float64[:]),
            ("param_limit_pos", float64[:]),
            ("param_limit_neg", float64[:]),
            ("inner_relative_tol", float64),
            ("flag_Ws_update", boolean),
            ("upper", boolean),
        )
        return spec

    def params_to_dict(self) -> dict:
        return dict(
            iter_limit=self.iter_limit,
            relative_tol=self.relative_tol,
            workingset=self.workingset,
            dualpruning=self.dualpruning,
            screening=self.screening,
            simpruning=self.simpruning,
        )

    def setup(
        self,
        datafit: Union[BaseDatafit, JitClassType],
        penalty: Union[BasePenalty, JitClassType],
        A: NDArray,
        lmbd: float,
    ) -> None:
        """Setup the bounding solver with problem data."""
        self.lipschitz = datafit.gradient_lipschitz_constant()
        self.A_colnorm = np.sqrt(np.sum(A**2, axis=0))
        self.stepsize = 1.0 / (self.lipschitz * self.A_colnorm**2)
        self.param_slope_pos = [
            penalty.param_slope_pos(i, lmbd) for i in range(A.shape[1])
        ]
        self.param_slope_neg = np.array(
            [penalty.param_slope_neg(i, lmbd) for i in range(A.shape[1])]
        )
        self.param_limit_pos = np.array(
            [penalty.param_limit_pos(i, lmbd) for i in range(A.shape[1])]
        )
        self.param_limit_neg = np.array(
            [penalty.param_limit_neg(i, lmbd) for i in range(A.shape[1])]
        )

    def bound(
        self,
        datafit: Union[BaseDatafit, JitClassType],
        penalty: Union[BasePenalty, JitClassType],
        A: NDArray,
        lmbd: float,
        S0: NDArray,
        S1: NDArray,
        Sb: NDArray,
        x: NDArray,
        ub: float,
        upper: bool,
    ):
        """Solve the bounding problem."""

        # ----- Initialize working values ----- #

        self.S0 = S0
        self.S1 = S1
        self.Sb = Sb
        self.Sbi = np.copy(Sb)

        if self.workingset and not upper:
            self.Ws = self.S1 | (x != 0.0)
        else:
            self.Ws = self.S1 | self.Sbi

        self.x = x
        self.w = np.dot(A[:, self.Ws], x[self.Ws])
        self.u = -datafit.gradient(self.w)
        self.v = np.empty_like(x)

        self.pv = np.inf
        self.pv_old = np.inf
        self.dv = np.nan
        self.ub = ub

        self.inner_relative_tol = 0.1 * self.relative_tol
        self.flag_Ws_update = False
        self.upper = upper

        while True:

            self.v = np.empty_like(x)

            for _ in range(self.iter_limit):

                self.pv_old = self.pv
                self.inner_loop(datafit, penalty, A)
                self.update_pv(datafit, penalty, lmbd)
                if self.inner_stopping_criterion():
                    break

            self.workingset_update(A)
            self.update_dv(datafit, penalty, A, lmbd)

            if upper or self.outer_stopping_criterion():
                break

            if self.dualpruning:
                if self.dv > ub:
                    break
            if self.screening:
                self.screening_tests(datafit, A)
            if self.simpruning:
                self.simpruning_tests(datafit, penalty, A, lmbd)

    def update_pv(
        self,
        datafit: Union[BaseDatafit, JitClassType],
        penalty: Union[BasePenalty, JitClassType],
        lmbd: float,
    ):
        pv = datafit.value(self.w)
        for i in np.flatnonzero(self.S1):
            pv += lmbd + penalty.value(i, self.x[i])
        for i in np.flatnonzero(self.Sbi):
            if 0.0 <= self.x[i] <= self.param_limit_pos[i]:
                pv += self.param_slope_pos[i] * self.x[i]
            elif 0.0 >= self.x[i] >= self.param_limit_neg[i]:
                pv += self.param_slope_neg[i] * self.x[i]
            else:
                pv += lmbd + penalty.value(i, self.x[i])
        self.pv = pv

    def update_dv(
        self,
        datafit: Union[BaseDatafit, JitClassType],
        penalty: Union[BasePenalty, JitClassType],
        A: NDArray,
        lmbd: float,
    ):
        dv = -datafit.conjugate(-self.u)
        for i in np.flatnonzero(self.S1):
            self.v[i] = np.dot(A[:, i], self.u)
            dv -= penalty.conjugate(i, self.v[i]) - lmbd
        for i in np.flatnonzero(self.Sbi):
            self.v[i] = np.dot(A[:, i], self.u)
            dv -= np.maximum(penalty.conjugate(i, self.v[i]) - lmbd, 0.0)
        self.dv = dv

    def inner_loop(
        self,
        datafit: Union[BaseDatafit, JitClassType],
        penalty: Union[BasePenalty, JitClassType],
        A: NDArray,
    ):
        for i in np.flatnonzero(self.Ws):
            ai = A[:, i]
            xi = self.x[i]
            ci = xi + self.stepsize[i] * np.dot(ai, self.u)
            if self.Sb[i]:
                sni = self.stepsize[i] * self.param_slope_neg[i]
                spi = self.stepsize[i] * self.param_slope_pos[i]
                if sni <= ci <= spi:
                    self.x[i] = 0.0
                elif sni + self.param_limit_neg[i] <= ci < sni:
                    self.x[i] = ci - sni
                elif spi + self.param_limit_pos[i] >= ci > spi:
                    self.x[i] = ci - spi
                else:
                    self.x[i] = penalty.prox(i, ci, self.stepsize[i])
            elif self.S1[i]:
                self.x[i] = penalty.prox(i, ci, self.stepsize[i])
            if self.x[i] != xi:
                self.w += (self.x[i] - xi) * ai
                self.u = -datafit.gradient(self.w)

    def inner_stopping_criterion(self):
        if (
            np.abs(self.pv - self.pv_old) / np.abs(self.pv)
            <= self.inner_relative_tol
        ):
            return True
        return False

    def outer_stopping_criterion(self):
        if not self.flag_Ws_update:
            if (
                np.abs(self.pv - self.dv) / np.abs(self.pv)
                <= self.relative_tol
            ):
                return True
            if self.inner_relative_tol <= 1e-12:
                return True
            self.inner_relative_tol *= 1e-2
        return False

    def workingset_update(self, A: NDArray):
        self.flag_Ws_update = False
        for i in np.flatnonzero(np.logical_not(self.Ws) & self.Sbi):
            self.v[i] = np.dot(A[:, i], self.u)
            if (
                self.v[i] >= self.param_slope_pos[i]
                or self.v[i] <= self.param_slope_neg[i]
            ):
                self.Ws[i] = True
                self.flag_Ws_update = True

    def screening_tests(
        self, datafit: Union[BaseDatafit, JitClassType], A: NDArray
    ):
        flag = False
        r = np.sqrt(2.0 * np.abs(self.pv - self.dv) * self.lipschitz)
        for i in np.flatnonzero(self.Sbi):
            r_neg = self.param_slope_neg[i] + r * self.A_colnorm[i]
            r_pos = self.param_slope_pos[i] - r * self.A_colnorm[i]
            if r_neg < self.v[i] < r_pos:
                if self.x[i] != 0.0:
                    self.w -= self.x[i] * A[:, i]
                    self.x[i] = 0.0
                    flag = True
                self.Ws[i] = False
                self.Sbi[i] = False
        if flag:
            self.u = -datafit.gradient(self.w)

    def simpruning_tests(
        self,
        datafit: Union[BaseDatafit, JitClassType],
        penalty: Union[BasePenalty, JitClassType],
        A: NDArray,
        lmbd: float,
    ):
        flag_update = False
        flag_passed = True
        while flag_passed:
            flag_passed = False
            for i in np.flatnonzero(self.Sb):
                p = penalty.conjugate(i, self.v[i]) - lmbd
                pp = np.maximum(p, 0.0)
                pn = np.maximum(-p, 0.0)
                if self.dv + pn > self.ub:
                    flag_passed = True
                    self.Sb[i] = False
                    self.S0[i] = True
                    self.Ws[i] = False
                    self.Sbi[i] = False
                    if self.x[i] != 0.0:
                        self.w -= self.x[i] * A[:, i]
                        self.x[i] = 0.0
                        flag_update = True
                    self.dv += pp
                elif self.dv + pp > self.ub:
                    flag_passed = True
                    self.Sb[i] = False
                    self.S1[i] = True
                    self.Ws[i] = True
                    self.Sbi[i] = False
                    self.dv += pn
        if flag_update:
            self.u = -datafit.gradient(self.w)


class ProblemWrapper(pb.Problem):
    """`pybnb <https://pybnb.readthedocs.io/en/stable/>`_ wrapper for
    L0-regularized problems."""

    def __init__(
        self,
        datafit: Union[BaseDatafit, JitClassType],
        penalty: Union[BaseDatafit, JitClassType],
        A: NDArray,
        lmbd: float,
        x_init: Optional[NDArray] = None,
        bound_solver: Union[BoundSolver, JitClassType] = BoundSolver(),
    ):

        # Prepare problem data
        if isinstance(datafit, CompilableClass):
            if not isinstance(datafit, JitClassType):
                datafit = compiled_clone(datafit)
        if isinstance(penalty, CompilableClass):
            if not isinstance(penalty, JitClassType):
                penalty = compiled_clone(penalty)
        if isinstance(datafit, JitClassType):
            if isinstance(penalty, JitClassType):
                if not isinstance(self.bound_solver, JitClassType):
                    self.bound_solver = compiled_clone(self.bound_solver)
        if not A.flags.f_contiguous:
            A = np.array(A, order="F")
        if x_init is None:
            x_init = np.zeros(A.shape[1])

        # Problem data
        self.datafit = datafit
        self.penalty = penalty
        self.A = A
        self.lmbd = lmbd
        self.m, self.n = A.shape

        # Solver state
        self.tree_state = TreeState(np.inf)

        # Node state
        self.node_state = NodeState(
            -1,
            np.zeros(self.n, dtype=bool),
            np.zeros(self.n, dtype=bool),
            np.ones(self.n, dtype=bool),
            x_init,
            np.zeros(self.n),
            -np.inf,
            np.inf,
        )

        # Bounding solver
        self.bound_solver = bound_solver
        self.bound_solver.setup(datafit, penalty, A, lmbd)

        # Solver trace
        self.trace = dict()

    def sense(self):
        return pb.minimize

    def objective(self) -> float:

        if self.node_state.category == 0:
            return self.node_state.upper_bound

        if not np.any(self.node_state.Sb):
            self.node_state.x_upper = np.copy(self.node_state.x_lower)
            self.node_state.upper_bound = self.node_state.lower_bound
            return self.node_state.upper_bound

        if np.all(
            (self.node_state.x_lower[self.node_state.Sb] == 0.0)
            | (
                self.node_state.x_lower[self.node_state.Sb]
                >= self.bound_solver.param_limit_pos[self.node_state.Sb]
            )
            | (
                self.node_state.x_lower[self.node_state.Sb]
                <= self.bound_solver.param_limit_neg[self.node_state.Sb]
            )
        ):
            self.node_state.x_upper = np.copy(self.node_state.x_lower)
            self.node_state.upper_bound = self.node_state.lower_bound
            return self.node_state.upper_bound

        self.bound_solver.bound(
            self.datafit,
            self.penalty,
            self.A,
            self.lmbd,
            self.node_state.S0 | self.node_state.Sb,
            self.node_state.S1,
            np.zeros(self.n, dtype=bool),
            self.node_state.x_upper,
            self.tree_state.best_upper_bound,
            True,
        )

        self.node_state.x_upper = self.bound_solver.x
        self.node_state.upper_bound = self.bound_solver.pv

        return self.node_state.upper_bound

    def bound(self) -> float:

        self.bound_solver.bound(
            self.datafit,
            self.penalty,
            self.A,
            self.lmbd,
            self.node_state.S0,
            self.node_state.S1,
            self.node_state.Sb,
            self.node_state.x_lower,
            self.tree_state.best_upper_bound,
            False,
        )

        self.node_state.S0 = self.bound_solver.S0
        self.node_state.S1 = self.bound_solver.S1
        self.node_state.Sb = self.bound_solver.Sb
        self.node_state.x_lower = self.bound_solver.x
        self.node_state.lower_bound = self.bound_solver.dv

        return self.node_state.lower_bound

    def save_state(self, node: pb.Node) -> None:
        node.state = self.node_state

    def load_state(self, node: pb.Node) -> None:
        self.node_state = node.state

    def branch(self):

        if not np.any(self.node_state.Sb):
            return

        # TODO: make the two following lines more readable
        jSb = np.argmax(np.abs(self.node_state.x_lower[self.node_state.Sb]))
        j = np.arange(self.n)[self.node_state.Sb][jSb]

        child = pb.Node()
        child.state = deepcopy(self.node_state)
        child.state.fix_to(j, 0)
        yield child

        child = pb.Node()
        child.state = deepcopy(self.node_state)
        child.state.fix_to(j, 1)
        yield child

    def notify_solve_begins(
        self,
        comm,
        worker_comm,
        convergence_checker,
    ):
        pass

    def notify_new_best_node(
        self,
        node: pb.Node,
        current: bool,
    ):
        if current:
            self.tree_state.best_upper_bound = self.node_state.upper_bound

    def notify_solve_finished(
        self,
        comm,
        worker_comm,
        results,
    ):
        pass


class BnbSolver:
    r"""Branch-and-Bound solver for L0-regularized problems.

    The problem is expressed as

    .. math::

        \textstyle\min_{\mathbf{x} \in \mathbb{R}^{n}} f(\mathbf{Ax}) + \lambda\|\mathbf{x}\|_0 + h(\mathbf{x})

    where :math:`f` is a :class:`el0ps.datafit.BaseDatafit` function,
    :math:`\mathbf{A} \in \mathbb{R}^{m \times n}` is a matrix, :math:`h` is a
    :class:`el0ps.penalty.BasePenalty` function, and :math:`\lambda` is a
    positive scalar.

    Parameters
    ----------
    relative_gap: float, default=1e-8
        Relative gap targeted on the objective value by the solver.
    absolute_gap: float, default=0.0
        Absolute gap targeted on the objective value by the solver.
    time_limit: float, default=None
        Limit in second on the solving time.
    node_limit: int, default=None
        Limit on the number of nodes explored during the BnB algorithm.
    queue_limit: int, default=None
        Limit on the number of nodes in the queue during the BnB algorithm.
    queue_strategy: str, default="bound"
        Queue strategy during the BnB algorithm. Can be one of the following:
        - "bound": select nodes with the worst lower bound first
        - "objective": select nodes with the best upper bound first
        - "breadth": select shallower nodes in the BnB tree first
        - "depth": select deepest nodes in the BnB tree first
        - "fifo": select nodes lastly added to the queue first
        - "lifo": select nodes firstly added to the queue first
        - "random": select nodes randomly in the queue
        - "local_gap": select nodes with the closest gap between their upper
        and lower bounds first
    verbose: bool, default=True
        Toggle solver verbosity.
    **kwargs: keyword arguments
        Additional keyword arguments passed to a :class:`BoundSolver` instance
        used for the bounding step of the Branch-and-Bound algorithm.
    """  # noqa: E501

    def __init__(
        self,
        relative_gap: float = 1e-8,
        absolute_gap: float = 0.0,
        time_limit: float = np.inf,
        node_limit: Optional[int] = None,
        queue_limit: Optional[int] = None,
        queue_strategy: str = "bound",
        verbose: bool = False,
        **kwargs,
    ):
        self.solver = pb.Solver()
        self.relative_gap = relative_gap
        self.absolute_gap = absolute_gap
        self.node_limit = node_limit
        self.time_limit = time_limit
        self.queue_limit = queue_limit
        self.queue_strategy = queue_strategy
        self.verbose = verbose

        self.bound_solver = BoundSolver(**kwargs)

    @property
    def accept_jitclass(self) -> bool:
        return True

    def package_results(
        self,
        problem: pb.Problem,
        results: pb.SolverResults,
    ) -> Result:

        status = Status.UNKNOWN
        if results.solution_status == pb.SolutionStatus.optimal:
            status = Status.OPTIMAL
        elif results.solution_status == pb.SolutionStatus.invalid:
            absolute_gap = np.abs(results.objective - results.bound)
            relative_gap = np.abs(results.objective - results.bound) / max(
                1.0, np.abs(results.objective)
            )
            if (
                absolute_gap <= self.absolute_gap
                or relative_gap <= self.relative_gap
            ):
                status = Status.OPTIMAL
            if (
                results.termination_condition
                == pb.TerminationCondition.optimality
            ):
                status = Status.OPTIMAL
        elif results.solution_status == pb.SolutionStatus.feasible:
            if (
                results.termination_condition
                == pb.TerminationCondition.node_limit
            ):
                status = Status.ITER_LIMIT
            elif (
                results.termination_condition
                == pb.TerminationCondition.time_limit
            ):
                status = Status.TIME_LIMIT
            elif (
                results.termination_condition
                == pb.TerminationCondition.queue_limit
            ):
                status = Status.MEMORY_LIMIT
            else:
                status = Status.OPTIMAL
        elif results.solution_status == pb.SolutionStatus.infeasible:
            status = Status.INFEASIBLE
        elif results.solution_status == pb.SolutionStatus.unbounded:
            status = Status.UNBOUNDED

        return Result(
            status,
            results.wall_time,
            results.nodes,
            results.best_node.state.x_upper,
            results.objective,
            problem.trace,
        )

    def solve(
        self,
        datafit: Union[BaseDatafit, CompilableClass, JitClassType],
        penalty: Union[BaseDatafit, CompilableClass, JitClassType],
        A: NDArray,
        lmbd: float,
        x_init: Optional[NDArray] = None,
    ):
        """Solve an L0-regularized problem.

        Parameters
        ----------
        datafit: Union[BaseDatafit, JitClassType]
            Problem datafit function. If not already JIT-compiled the solver
            automatically compiles if it derives from the
            :class:`el0ps.compilation.CompilableClass`.
        penalty: Union[BaseDatafit, JitClassType]
            Problem penalty function. If not already JIT-compiled the solver
            automatically compiles if it derives from the
            :class:`el0ps.compilation.CompilableClass`.
        A: NDArray
            Problem matrix.
        lmbd: float
            Problem L0-norm weight parameter.
        x_init: NDArray, default=None
            Initial point for the solver. If `None`, the solver initializes it
            to the all-zero vector.
        """

        problem = ProblemWrapper(
            datafit, penalty, A, lmbd, x_init, self.bound_solver
        )

        results = self.solver.solve(
            problem,
            absolute_gap=self.absolute_gap,
            relative_gap=self.relative_gap,
            node_limit=self.node_limit,
            time_limit=self.time_limit,
            queue_limit=self.queue_limit,
            queue_strategy=self.queue_strategy,
            log=pb.solver._notset if self.verbose else None,
        )

        return self.package_results(problem, results)



import numpy as np
import pytest

from el0ps.datafit import (
    BaseDatafit,
    KullbackLeibler,
    Leastsquares,
    Logcosh,
    Logistic,
    Squaredhinge,
)

m = 100
y = np.random.randn(m)
x = np.random.randn(m)
u = np.random.randn(m)
datafits = [
    pytest.param(datafit, id=f"{datafit.__class__.__name__}")
    for datafit in [
        KullbackLeibler(np.abs(y)),
        Leastsquares(y),
        Logcosh(y),
        Logistic(2.0 * (y > 0.0) - 1.0),
        Squaredhinge(2.0 * (y > 0.0) - 1.0),
    ]
]


@pytest.mark.parametrize("datafit", datafits)
def test_datafit(datafit: BaseDatafit):
    assert isinstance(datafit.__str__(), str)
    assert datafit.value(x) + datafit.conjugate(u) >= np.dot(x, u)
    assert datafit.gradient_lipschitz_constant() >= 0.0
    assert datafit.gradient(x).shape == x.shape
    assert datafit.value(x) >= datafit.value(u) + np.dot(
        datafit.gradient(u), x - u
    )



"""Datafit functions."""

from .base import BaseDatafit, MipDatafit
from .kullbackleibler import KullbackLeibler
from .leastsquares import Leastsquares
from .logcosh import Logcosh
from .logistic import Logistic
from .squaredhinge import Squaredhinge

__all__ = [
    "BaseDatafit",
    "MipDatafit",
    "KullbackLeibler",
    "Leastsquares",
    "Logcosh",
    "Logistic",
    "Squaredhinge",
]



"""Compilation utilities."""

from abc import abstractmethod
from functools import lru_cache
from numba.experimental import jitclass


class CompilableClass:
    """Template for classes that can be compiled with Numba."""

    @abstractmethod
    def get_spec(self) -> tuple:
        """Specify the numba types of the class attributes.

        Returns
        -------
        spec: Tuple of (attr_name, dtype)
            Specs to be passed to Numba jitclass to compile the class.
        """
        ...

    @abstractmethod
    def params_to_dict(self) -> dict:
        """Returns the parameters name and value used to initialize the class
        instance.

        Returns
        -------
        dict_of_params: dict
            The parameters name and value used to initialize the class
            instance.
        """
        ...


@lru_cache()
def compiled_clone(instance: CompilableClass):
    r"""Compile a class instance to a
    `jitclass <https://numba.readthedocs.io/en/stable/user/jitclass.html>`_.
    This function is inspired from a similar one in the
    `skglm <https://github.com/scikit-learn-contrib/skglm/tree/main>`_ package.

    Parameters
    ----------
    instance: object
        Instance to compile.

    Returns
    -------
    compiled_instance: jitclass
        Compiled instance.
    """
    cls = instance.__class__
    spec = instance.get_spec()
    params = instance.params_to_dict()
    return jitclass(spec)(cls)(**params)



import numpy as np
from numba import float64
from numpy.typing import NDArray
from el0ps.compilation import CompilableClass

from el0ps.datafit.base import BaseDatafit


class KullbackLeibler(CompilableClass, BaseDatafit):
    r"""Kullback-Leibler datafit function.

    The function is defined as

    .. math::

        f(\mathbf{w}) = \sum_{i=1}^m y_i \log(\tfrac{y_i}{w_i + e}) + (w_i + e) - y_i


    for some :math:`\mathbf{y} \in \mathbb{R}^m` and :math:`e > 0`.

    Parameters
    ----------
    y : NDArray
        Data vector.
    e : float = 1e-8
        Smoothing parameter.
    """  # noqa: E501

    def __init__(self, y: NDArray, e: float = 1e-8) -> None:
        self.y = y
        self.e = e
        self.L = np.max(y) / e**2
        self.log_yy = np.log(y * y)

    def __str__(self) -> str:
        return "KullbackLeibler"

    def get_spec(self) -> tuple:
        spec = (
            ("y", float64[::1]),
            ("e", float64),
            ("L", float64),
            ("log_yy", float64[::1]),
        )
        return spec

    def params_to_dict(self) -> dict:
        return dict(y=self.y, e=self.e)

    def value(self, w: NDArray) -> float:
        z = np.maximum(w, 0.0) + self.e
        return np.sum(self.y * np.log(self.y / z) + z - self.y)

    def conjugate(self, w: NDArray) -> float:
        u = w
        v = 1.0 - u
        if np.any(v <= 0.0):
            return np.inf
        return np.sum(self.y * (self.log_yy - np.log(v)) - self.e * u)

    def gradient(self, w: NDArray) -> NDArray:
        z = np.maximum(w, 0.0) + self.e
        return 1.0 - self.y / z

    def gradient_lipschitz_constant(self) -> float:
        return self.L



import numpy as np
import pyomo.kernel as pmo
from numba import float64
from numpy.typing import NDArray

from el0ps.compilation import CompilableClass
from el0ps.datafit.base import BaseDatafit, MipDatafit


class Squaredhinge(CompilableClass, BaseDatafit, MipDatafit):
    r"""Squared-hinge datafit function.

    The function is defined as

    .. math::

        f(\mathbf{w}) = \sum_{i=1}^m \max(1 - y_i w_i, 0)^2

    for some :math:`\mathbf{y} \in \mathbb{R}^m`.

    Parameters
    ----------
    y : NDArray
        Data vector.
    """

    def __init__(self, y: NDArray) -> None:
        self.y = y
        self.L = 2.0

    def __str__(self) -> str:
        return "Squaredhinge"

    def get_spec(self) -> tuple:
        spec = (("y", float64[::1]), ("L", float64))
        return spec

    def params_to_dict(self) -> dict:
        return dict(y=self.y)

    def value(self, w: NDArray) -> float:
        v = np.maximum(1.0 - self.y * w, 0.0)
        return np.dot(v, v)

    def conjugate(self, w: NDArray) -> float:
        v = np.maximum(-0.5 * (self.y * w), 0.0)
        return 0.5 * np.dot(w, w) + np.dot(self.y, w) - np.dot(v, v)

    def gradient(self, w: NDArray) -> NDArray:
        return -2.0 * self.y * np.maximum(1.0 - self.y * w, 0.0)

    def gradient_lipschitz_constant(self) -> float:
        return self.L

    def bind_model(self, model: pmo.block) -> None:
        model.f1_var = pmo.variable_dict()
        model.f1_con = pmo.constraint_dict()
        for j in model.M:
            model.f1_var[j] = pmo.variable(domain=pmo.NonNegativeReals)
            model.f1_con[j] = pmo.constraint(
                model.f1_var[j] >= 1.0 - self.y[j] * model.w[j]
            )
        model.c_var = pmo.variable()
        model.c_con = pmo.constraint(model.c_var == 0.5)
        model.f_con = pmo.conic.rotated_quadratic(
            model.f,
            model.c_var,
            [model.f1_var[j] for j in model.M],
        )



import numpy as np
from numpy.typing import NDArray
from numba import float64

from el0ps.compilation import CompilableClass
from el0ps.penalty.base import SymmetricPenalty


class L1norm(CompilableClass, SymmetricPenalty):
    r"""L1-norm :class:`BasePenalty` penalty function.

    The splitting terms are expressed as

    .. math::
        h_i(x_i) = \alpha|x_i|

    for some :math:`\alpha > 0`.

    Parameters
    ----------
    alpha: float
        L1-norm weight.
    """

    def __init__(self, alpha: float) -> None:
        self.alpha = alpha

    def __str__(self) -> str:
        return "L1norm"

    def get_spec(self) -> tuple:
        spec = (("alpha", float64),)
        return spec

    def params_to_dict(self) -> dict:
        return dict(alpha=self.alpha)

    def value(self, i: int, x: float) -> float:
        return self.alpha * np.abs(x)

    def conjugate(self, i: int, x: float) -> float:
        return 0.0 if np.abs(x) <= self.alpha else np.inf

    def prox(self, i: int, x: float, eta: float) -> float:
        return np.sign(x) * np.maximum(0.0, np.abs(x) - eta * self.alpha)

    def subdiff(self, i: int, x: float) -> NDArray:
        if x == 0:
            return [-self.alpha, self.alpha]
        else:
            s = self.alpha * np.sign(x)
            return [s, s]

    def conjugate_subdiff(self, i: int, x: float) -> NDArray:
        if np.abs(x) < self.alpha:
            return [0.0, 0.0]
        elif x == -self.alpha:
            return [-np.inf, 0.0]
        elif x == self.alpha:
            return [0.0, np.inf]
        else:
            return [np.nan, np.nan]

    def param_slope(self, i: int, lmbd: float) -> float:
        return self.alpha

    def param_limit(self, i: int, lmbd: float) -> float:
        return np.inf

    def param_bndry(self, i, lmbd):
        return np.inf



import numpy as np
import pytest
from pyomo.opt.base.solvers import check_available_solvers
from el0ps.datafit import Leastsquares
from el0ps.penalty import Bigm
from el0ps.utils import compute_lmbd_max
from el0ps.solver import Status, BnbSolver, MipSolver, OaSolver
from el0ps.solver.mip import _mip_optim_bindings
from .utils import make_regression


A, y, x_true = make_regression(3, 15, 20)
M = 1.5 * np.max(np.abs(x_true))
datafit = Leastsquares(y)
penalty = Bigm(M)
lmbd = 0.1 * compute_lmbd_max(datafit, penalty, A)
x_init = np.zeros(x_true.shape)

solvers = [
    pytest.param(solver, id=f"{solver.__class__.__name__}")
    for solver in [
        BnbSolver(),
        MipSolver(optimizer_name="cplex"),
        MipSolver(optimizer_name="gurobi"),
        MipSolver(optimizer_name="mosek"),
        OaSolver(optimizer_name="cplex"),
        OaSolver(optimizer_name="gurobi"),
        OaSolver(optimizer_name="mosek"),
    ]
]


@pytest.mark.parametrize("solver", solvers)
def test_solver(solver):

    if isinstance(solver, (MipSolver, OaSolver)):
        if not check_available_solvers(
            _mip_optim_bindings[solver.optimizer_name]["optimizer_name"]
        ):
            pytest.skip(f"{solver.optimizer_name} not available")

    result = solver.solve(datafit, penalty, A, lmbd, x_init=x_init)
    assert result.status == Status.OPTIMAL



"""Base classes for L0-penalized problem solvers and related utilities."""

from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum
from numpy.typing import NDArray
from typing import Optional
from el0ps.datafit import BaseDatafit
from el0ps.penalty import BasePenalty


class Status(Enum):
    """Solver status.

    Attributes
    ----------
    UNKNOWN: str
        Unknown solver status.
    ITER_LIMIT: str
        The solver reached the iteration limit.
    TIME_LIMIT: str
        The solver reached the time limit.
    MEMORY_LIMIT: str
        The solver reached the memory limit.
    RUNNING: str
        The solver is running.
    INFEASIBLE: str
        The problem is infeasible.
    UNBOUNDED: str
        The problem is unbounded.
    OPTIMAL: str
        The solver found an optimal solution.
    """

    def __str__(self):
        return str(self.value)

    UNKNOWN = "unknown"
    ITER_LIMIT = "iter_limit"
    TIME_LIMIT = "time_limit"
    MEMORY_LIMIT = "memory_limit"
    RUNNING = "running"
    INFEASIBLE = "infeasible"
    UNBOUNDED = "unbounded"
    OPTIMAL = "optimal"


@dataclass
class Result:
    """Solver results.

    Attributes
    ----------
    status: Status
        Solver status.
    solve_time: float
        Solve time in seconds.
    iter_count: int
        Iteration count.
    x: NDArray
        Problem solution.
    objective_value: float
        Objective value.
    trace: dict, default=None
        Solver trace if available, otherwise `None`.
    """

    status: Status
    solve_time: float
    iter_count: int
    x: NDArray
    objective_value: float
    trace: Optional[dict]

    def __str__(self) -> str:
        s = ""
        s += "Result\n"
        s += "  Status     : {}\n".format(self.status.value)
        s += "  Solve time : {:f} seconds\n".format(self.solve_time)
        s += "  Iter count : {:d}\n".format(self.iter_count)
        s += "  Objective  : {:f}\n".format(self.objective_value)
        s += "  Non-zeros  : {:d}".format(sum(self.x != 0.0))
        return s


class BaseSolver:
    r"""Base class for solvers of L0-regularized problem.

    The problem is expressed as

    .. math::

        \textstyle\min_{\mathbf{x} \in \mathbb{R}^{n}} f(\mathbf{Ax}) + \lambda\|\mathbf{x}\|_0 + h(\mathbf{x})

    where :math:`f` is a :class:`el0ps.datafit.BaseDatafit` function,
    :math:`\mathbf{A} \in \mathbb{R}^{m \times n}` is a matrix, :math:`h` is a
    :class:`el0ps.penalty.BasePenalty` function, and :math:`\lambda` is a
    positive scalar."""  # noqa: E501

    @property
    def accept_jitclass(self) -> bool:
        """Return whether if the solver accepts a
        `jitclass <https://numba.readthedocs.io/en/stable/user/jitclass.html>`_
        for the datafit and penalty function. This method can be overridden in
        derived classes to provide a more specific implementation."""
        return False

    @abstractmethod
    def solve(
        self,
        datafit: BaseDatafit,
        penalty: BasePenalty,
        A: NDArray,
        lmbd: float,
        x_init: Optional[NDArray] = None,
    ) -> Result:
        r"""Solve an L0-regularized problem.

        Parameters
        ----------
        datafit: BaseDatafit
            Problem datafit function.
        penalty: BasePenalty
            Problem penalty function.
        A: NDArray
            Problem matrix.
        lmbd: float, positive
            Problem L0-norm weight.
        x_init: NDArray, default=None
            Stating point for the solver.

        Returns
        -------
        result: Result
            Solver results.
        """
        ...



import numpy as np
from typing import Optional, Union
from numba.experimental.jitclass.base import JitClassType
from numpy.typing import NDArray

from el0ps.compilation import compiled_clone
from el0ps.datafit import BaseDatafit
from el0ps.penalty import BasePenalty
from el0ps.solver import BaseSolver, Result, Status
from el0ps.utils import compute_lmbd_max


class Path:
    r"""Regularization path fitting for L0-regularized problems.

    The problem is expressed as

    .. math::

        \textstyle\min_{\mathbf{x} \in \mathbb{R}^{n}} f(\mathbf{Ax}) + \lambda\|\mathbf{x}\|_0 + h(\mathbf{x})

    where :math:`f` is a :class:`el0ps.datafit.BaseDatafit` function,
    :math:`\mathbf{A} \in \mathbb{R}^{m \times n}` is a matrix, :math:`h` is a
    :class:`el0ps.penalty.BasePenalty` function, and :math:`\lambda` is a
    positive scalar. The path fitting consists of solving this problem over a
    range of values of parameter :math:`\lambda`.

    Parameters
    ----------
    lmbds : list, default=None
        Values of parameter ``lmbd`` to consider. If ``None``, the values
        considered is computed from the other parameters ``lmbd_max``,
        ``lmbd_min``, ``lmbd_num``, ``lmbd_normalized``, and ``lmbd_spacing``.
    lmbd_max : float = 1e-0
        Maximum value of ``lmbd`` to consider. If ``lmbds`` is not ``None``,
        this parameter is ignored.
    lmbd_min : float = 1e-2
        Minimum value of ``lmbd`` to consider. If ``lmbds`` is not ``None``,
        this parameter is ignored.
    lmbd_num : int = 10
        Number of values of ``lmbd`` to consider. If ``lmbds`` is not ``None``,
        this parameter is ignored.
    lmbd_scale : str = "log"
        Scale of the values of ``lmbd`` to consider. Can be ``linear`` or
        ``log``. If ``lmbds`` is not ``None``, this parameter is ignored.
    lmbd_normalized : bool = True
        If ``True``, the values of ``lmbd`` considered are scaled by the
        value outputted by the function :func:`el0ps.utils.compute_lmbd_max`
        so that the solution to the problem is the all-zero vector for the
        largest value of ``lmbd`` specified in the regularization path.
        If ``lmbds`` is not `None`, this value is ignored.
    max_nnz : int = sys.maxsize
        Stop the path fitting when a solution with more than ``max_nnz``
        non-zero coefficients is found for a given value of ``lmbd``. When
        ``None``, this criterion is ignored.
    stop_if_not_optimal : bool = True
        Stop the path fitting when the problem at a given value of ``lmbd``
        is not solved to optimality.
    verbose : bool = True
        Toggle displays during path fitting.
    """  # noqa: E501

    def __init__(
        self,
        lmbds: Optional[list] = None,
        lmbd_max: float = 1e-0,
        lmbd_min: float = 1e-2,
        lmbd_num: int = 10,
        lmbd_scale: str = "log",
        lmbd_normalized: bool = True,
        max_nnz: Optional[int] = None,
        stop_if_not_optimal: bool = True,
        verbose: bool = True,
    ) -> None:
        self.lmbds = lmbds
        self.lmbd_max = lmbd_max
        self.lmbd_min = lmbd_min
        self.lmbd_num = lmbd_num
        self.lmbd_scale = lmbd_scale
        self.lmbd_normalized = lmbd_normalized
        self.max_nnz = max_nnz
        self.stop_if_not_optimal = stop_if_not_optimal
        self.verbose = verbose

    def _get_lmbd_grid(
        self,
        datafit: Union[BaseDatafit, JitClassType],
        penalty: Union[BasePenalty, JitClassType],
        A: NDArray,
    ) -> NDArray:

        if self.lmbds is not None:
            lmbd_grid = np.array(self.lmbds)
        elif self.lmbd_scale == "log":
            lmbd_grid = np.logspace(
                np.log10(self.lmbd_max), np.log10(self.lmbd_min), self.lmbd_num
            )
        elif self.lmbd_scale == "lin":
            lmbd_grid = np.linspace(
                self.lmbd_max, self.lmbd_min, self.lmbd_num
            )
        else:
            raise ValueError(
                "Invalid value for parameter `lmbd_scale`. Must be either "
                "'log' or 'lin'."
            )

        if self.lmbd_normalized:
            lmbd_max_grid = compute_lmbd_max(datafit, penalty, A)
            lmbd_grid *= lmbd_max_grid / np.max(lmbd_grid)

        return lmbd_grid

    def _display_path_head(self) -> None:
        s = "  {:>10}  {:>10}  {:>10}  {:>10}  {:>10}".format(
            "lambda",
            "status",
            "time",
            "objective",
            "num nnz",
        )
        print("-" * len(s))
        print(s)
        print("-" * len(s))

    def _display_path_info(self, lmbd: float, result: Result) -> None:
        print(
            "  {:>10.2f}  {:>10}  {:>10.4f}  {:>10.2f}  {:>10d}".format(
                lmbd,
                result.status,
                result.solve_time,
                result.objective_value,
                np.count_nonzero(result.x),
            )
        )

    def _display_path_foot(self) -> None:
        print("-" * 60)

    def fit(
        self,
        solver: BaseSolver,
        datafit: Union[BaseDatafit, JitClassType],
        penalty: Union[BasePenalty, JitClassType],
        A: NDArray,
    ) -> dict:
        """Construct the regularization path.

        Parameters
        ----------
        solver: BaseSolver
            Solver to use for the path fitting.
        datafit: BaseDatafit
            Problem datafit function.
        penalty: BasePenalty
            Problem penalty function.
        A: NDArray
            Problem matrix.

        Returns
        -------
        results: dict
            The path fitting results indexed by the values of `lmbd`.
        """

        if solver.accept_jitclass:
            if isinstance(datafit, BaseDatafit):
                datafit = compiled_clone(datafit)
            if isinstance(penalty, BasePenalty):
                penalty = compiled_clone(penalty)
        if not A.flags.f_contiguous:
            A = np.array(A, order="F")

        lmbd_grid = self._get_lmbd_grid(datafit, penalty, A)

        if self.verbose:
            self._display_path_head()

        x_init = None
        results = {}

        for lmbd in lmbd_grid:

            # Solve the problem for the current value of lmbd
            result = solver.solve(datafit, penalty, A, lmbd, x_init)

            # Store results
            results[lmbd] = result
            x_init = np.copy(result.x)

            # Displays
            if self.verbose:
                self._display_path_info(lmbd, result)

            # Termination criteria
            if self.stop_if_not_optimal and result.status != Status.OPTIMAL:
                break
            if self.max_nnz is not None:
                if np.count_nonzero(result.x) > self.max_nnz:
                    break

        if self.verbose:
            self._display_path_foot()

        return results



import numpy as np
from numba import float64
from numpy.typing import NDArray

from el0ps.compilation import CompilableClass
from el0ps.datafit.base import BaseDatafit


class Logcosh(CompilableClass, BaseDatafit):
    r"""Logcosh datafit function.

    The function is defined as

    .. math::

        f(\mathbf{w}) = \sum_{i=1}^m \log(\cosh(y_i - w_i))

    for some :math:`\mathbf{y} \in \mathbb{R}^m`.

    Parameters
    ----------
    y : NDArray
        Data vector.
    """

    def __init__(self, y: NDArray) -> None:
        self.y = y
        self.L = 1.0

    def __str__(self) -> str:
        return "Logcosh"

    def get_spec(self) -> tuple:
        spec = (("y", float64[::1]), ("L", float64))
        return spec

    def params_to_dict(self) -> dict:
        return dict(y=self.y)

    def value(self, w: NDArray) -> float:
        return np.sum(np.log(np.cosh(w - self.y)))

    def conjugate(self, w: NDArray) -> float:
        if np.max(np.abs(w)) > 1.0:
            return np.inf
        else:
            z = np.arctanh(w) + self.y
        return np.sum(z * w) - np.sum(np.log(np.cosh(z - self.y)))

    def gradient(self, w: NDArray) -> NDArray:
        return np.tanh(w - self.y)

    def gradient_lipschitz_constant(self) -> float:
        return self.L



import numpy as np
import pyomo.kernel as pmo
from numpy.typing import NDArray
from numba import float64

from el0ps.compilation import CompilableClass
from el0ps.penalty.base import SymmetricPenalty, MipPenalty


class L1L2norm(CompilableClass, SymmetricPenalty, MipPenalty):
    r"""L1L2-norm :class:`BasePenalty` penalty function.

    The splitting terms are expressed as

    .. math::
        h_i(x_i) = \alpha|x_i| + \beta x_i^2

    for some :math:`\alpha > 0` and :math:`\beta > 0`.

    Parameters
    ----------
    alpha: float
        L1-norm weight.
    beta: float
        L2-norm weight.
    """

    def __init__(self, alpha: float, beta: float) -> None:
        self.alpha = alpha
        self.beta = beta

    def __str__(self) -> str:
        return "L1L2norm"

    def get_spec(self) -> tuple:
        spec = (
            ("alpha", float64),
            ("beta", float64),
        )
        return spec

    def params_to_dict(self) -> dict:
        return dict(alpha=self.alpha, beta=self.beta)

    def value(self, i: int, x: float) -> float:
        return self.alpha * np.abs(x) + self.beta * x**2

    def conjugate(self, i: int, x: float) -> float:
        return np.maximum(np.abs(x) - self.alpha, 0.0) ** 2 / (4.0 * self.beta)

    def prox(self, i: int, x: float, eta: float) -> float:
        v = np.sign(x) / (1.0 + 2.0 * eta * self.beta)
        return v * np.maximum(np.abs(x) - eta * self.alpha, 0.0)

    def subdiff(self, i: int, x: float) -> NDArray:
        if x == 0:
            return [-self.alpha, self.alpha]
        else:
            s = self.alpha * np.sign(x) + 2.0 * self.beta * x
            return [s, s]

    def conjugate_subdiff(self, i: int, x: float) -> NDArray:
        if np.abs(x) < self.alpha:
            return [0.0, 0.0]
        elif x == -self.alpha:
            return [(x + self.alpha) / (2.0 * self.beta), 0.0]
        elif x == self.alpha:
            return [0.0, (x - self.alpha) / (2.0 * self.beta)]
        else:
            s = (x - self.alpha * np.sign(x)) / (2.0 * self.beta)
            return [s, s]

    def param_slope(self, i: int, lmbd: float) -> float:
        return self.alpha + np.sqrt(4.0 * self.beta * lmbd)

    def param_limit(self, i: int, lmbd: float) -> float:
        return np.sqrt(lmbd / self.beta)

    def param_bndry(self, i, lmbd):
        return (self.alpha + 2.0 * self.beta) * np.sqrt(lmbd / self.beta)

    def bind_model(self, model: pmo.block) -> None:

        model.h1_var = pmo.variable_dict()
        model.h2_var = pmo.variable_dict()
        for i in model.N:
            model.h1_var[i] = pmo.variable(domain=pmo.NonNegativeReals)
            model.h2_var[i] = pmo.variable(domain=pmo.NonNegativeReals)

        model.h1pos_con = pmo.constraint_dict()
        model.h1neg_con = pmo.constraint_dict()
        model.h2_con = pmo.constraint_dict()
        for i in model.N:
            model.h1pos_con[i] = pmo.constraint(model.h1_var[i] >= model.x[i])
            model.h1neg_con[i] = pmo.constraint(model.h1_var[i] >= -model.x[i])
            model.h2_con[i] = pmo.conic.rotated_quadratic(
                model.h2_var[i], model.z[i], [model.x[i]]
            )
        model.h_con = pmo.constraint(
            model.h
            >= (
                self.alpha * sum(model.h1_var[i] for i in model.N)
                + 2.0 * self.beta * sum(model.h2_var[i] for i in model.N)
            )
        )



"""Base classes for L0-norm classifier estimators."""

import numpy as np
from sklearn.linear_model._base import LinearClassifierMixin

from el0ps.solver import BaseSolver, BnbSolver
from el0ps.datafit import Squaredhinge
from el0ps.estimator.base import L0Estimator
from el0ps.estimator.utils import select_bigml1l2_penalty


class L0L1L2SVC(L0Estimator, LinearClassifierMixin):
    r"""Scikit-learn-compatible `linear model <https://scikit-learn.org/stable/api/sklearn.linear_model.html>`_
    support-vector-classifier estimators with L0L1L2-regularization.

    The estimator corresponds to a solution of the problem

    .. math::

        \textstyle\min_{\|\mathbf{x}\|_{\infty} \leq M} f(\mathbf{Ax}) + \lambda\|\mathbf{x}\|_0 + \alpha\|\mathbf{x}\|_1 + \beta\|\mathbf{x}\|_2^2

    where :math:`f` is a :class:`el0ps.datafit.Squaredhinge` function,
    :math:`\mathbf{A} \in \mathbb{R}^{m \times n}` is a matrix,
    :math:`\lambda > 0` is a parameter, the L0-norm :math:`\|\cdot\|_0` counts
    the number of non-zero entries in its input, and :math:`h` is a penalty
    function."""  # noqa: E501

    def __init__(
        self,
        lmbd: float,
        alpha: float,
        beta: float,
        M: float = np.inf,
        fit_intercept: bool = False,
        solver: BaseSolver = BnbSolver(),
    ):
        self.lmbd = lmbd
        self.alpha = alpha
        self.beta = beta
        self.M = M
        datafit = Squaredhinge(np.zeros(0))
        penalty = select_bigml1l2_penalty(alpha, beta, M)
        super().__init__(datafit, penalty, lmbd, fit_intercept, solver)


class L0SVC(L0L1L2SVC):
    """Substitute for :class:`.L0L1L2SVC` with parameters ``alpha=0`` and
    ``beta=0``."""

    def __init__(
        self,
        lmbd: float,
        M: float,
        fit_intercept: bool = False,
        solver: BaseSolver = BnbSolver(),
    ):
        super().__init__(lmbd, 0.0, 0.0, M, fit_intercept, solver)


class L0L1SVC(L0L1L2SVC):
    """Substitute for :class:`L0L1L2SVC` with parameter ``beta=0``."""

    def __init__(
        self,
        lmbd: float,
        alpha: float,
        M: float = np.inf,
        fit_intercept: bool = False,
        solver: BaseSolver = BnbSolver(),
    ):
        super().__init__(lmbd, alpha, 0.0, M, fit_intercept, solver)


class L0L2SVC(L0L1L2SVC):
    """Substitute for :class:`L0L1L2SVC` with parameter ``alpha=0``."""

    def __init__(
        self,
        lmbd: float,
        beta: float,
        M: float = np.inf,
        fit_intercept: bool = False,
        solver: BaseSolver = BnbSolver(),
    ):
        super().__init__(lmbd, 0.0, beta, M, fit_intercept, solver)



"""Mixed-Integer Programming solver for L0-regularized problems."""

import numpy as np
import pyomo.environ as pyo
import pyomo.kernel as pmo
import sys
from typing import Optional
from numpy.typing import NDArray
from pyomo.opt import OptSolver
from pyomo.opt import SolverResults

from el0ps.datafit import MipDatafit
from el0ps.penalty import MipPenalty
from el0ps.solver import BaseSolver, Result, Status


_mip_optim_bindings = {
    "cplex": {
        "optimizer_name": "cplex_persistent",
        "relative_gap": "mip_tolerances_mipgap",
        "absolute_gap": "mip_tolerances_absmipgap",
        "node_limit": "mip_limits_nodes",
        "time_limit": "timelimit",
        "verbose": "mip_display",
    },
    "gurobi": {
        "optimizer_name": "gurobi_persistent",
        "relative_gap": "MIPGap",
        "absolute_gap": "MIPGapAbs",
        "node_limit": "NodeLimit",
        "time_limit": "TimeLimit",
        "verbose": "OutputFlag",
    },
    "mosek": {
        "optimizer_name": "mosek_persistent",
        "relative_gap": "dparam.mio_tol_rel_gap",
        "absolute_gap": "dparam.mio_tol_abs_gap",
        "node_limit": "dparam.mio_max_nodes",
        "time_limit": "dparam.mio_max_time",
        "verbose": "iparam.log",
    },
}


class MipSolver(BaseSolver):
    r"""Mixed-integer programming solver for L0-regularized problems.

    The problem is expressed as

    .. math::

        \textstyle\min_{\mathbf{x} \in \mathbb{R}^{n}} f(\mathbf{Ax}) + \lambda\|\mathbf{x}\|_0 + h(\mathbf{x})

    where :math:`f` is a :class:`el0ps.datafit.BaseDatafit` function,
    :math:`\mathbf{A} \in \mathbb{R}^{m \times n}` is a matrix, :math:`h` is a
    :class:`el0ps.penalty.BasePenalty` function, and :math:`\lambda` is a
    positive scalar. To use this solver, the optimizer specified in the
    ``optimizer_name`` parameter must be installed and accessible by
    `pyomo <https://pyomo.readthedocs.io/en/stable/>`_ which is the underlying
    library used to model the problem.

    Parameters
    ----------
    optimizer_name: str = "gurobi"
        Mixed-Integer Programming optimizer to use. Available options are
        "cplex", "gurobi", and "mosek".
    relative_gap: float, default=1e-8
        Relative tolerance on the objective value.
    absolute_gap: float, default=0.0
        Absolute tolerance on the objective value.
    time_limit: float, default=None
        Limit in second on the solving time.
    node_limit: int, default=None
        Limit on the number of nodes explored by the MIP solver.
    queue_limit: int, default=None
        Limit on the number of nodes in the queue in the MIP solver.
    verbose: bool, default=False
        Whether to toggle solver verbosity.
    """  # noqa: E501

    def __init__(
        self,
        optimizer_name: str = "gurobi",
        relative_gap: float = 1e-8,
        absolute_gap: float = 0.0,
        time_limit: Optional[float] = None,
        node_limit: Optional[int] = None,
        queue_limit: Optional[int] = None,
        verbose: bool = False,
    ) -> None:
        self.optimizer_name = optimizer_name
        self.node_limit = node_limit if node_limit is not None else sys.maxsize
        self.queue_limit = (
            queue_limit if queue_limit is not None else sys.maxsize
        )
        self.time_limit = time_limit if time_limit is not None else np.inf
        self.relative_gap = relative_gap
        self.absolute_gap = absolute_gap
        self.verbose = verbose

    def __str__(self):
        return "MipSolver"

    def initialize_optimizer(self) -> OptSolver:
        if self.optimizer_name in _mip_optim_bindings:
            bindings = _mip_optim_bindings[self.optimizer_name]
            optim: OptSolver = pyo.SolverFactory(bindings["optimizer_name"])
            optim.options[bindings["relative_gap"]] = self.relative_gap
            optim.options[bindings["absolute_gap"]] = self.absolute_gap
            # Raises errors with cplex
            # optim.options[bindings["node_limit"]] = self.node_limit
            optim.options[bindings["time_limit"]] = self.time_limit
            optim.options[bindings["verbose"]] = int(self.verbose)
        else:
            raise ValueError(
                "Solver {} not supported. Available ones are: {}".format(
                    self.optimizer_name,
                    _mip_optim_bindings.keys(),
                )
            )
        return optim

    def build_model(
        self,
        datafit: MipDatafit,
        penalty: MipPenalty,
        A: NDArray,
        lmbd: float,
    ) -> pmo.block:
        model = pmo.block()
        model.M = range(A.shape[0])
        model.N = range(A.shape[1])
        model.x = pmo.variable_dict()
        model.z = pmo.variable_dict()
        for i in model.N:
            model.x[i] = pmo.variable(domain=pmo.Reals)
            model.z[i] = pmo.variable(domain=pmo.Binary)
        model.w = pmo.variable_dict()
        for j in model.M:
            model.w[j] = pmo.variable(domain=pmo.Reals)
        model.f = pmo.variable(domain=pmo.Reals)
        model.h = pmo.variable(domain=pmo.Reals)

        model.w_con = pmo.constraint_dict()
        for j in model.M:
            model.w_con[j] = pmo.constraint(
                model.w[j] == sum(A[j, i] * model.x[i] for i in model.N)
            )

        datafit.bind_model(model)
        penalty.bind_model(model)

        model.obj = pmo.objective(
            model.f + lmbd * sum(model.z[i] for i in model.N) + model.h
        )
        return model

    def package_result(self, model: pmo.block, result: SolverResults):
        if result.solver.termination_condition == "optimal":
            status = Status.OPTIMAL
        elif result.solver.termination_condition == "maxIterations":
            status = Status.ITER_LIMIT
        elif result.solver.termination_condition == "maxTimeLimit":
            status = Status.TIME_LIMIT
        elif result.solver.termination_condition == "unbounded":
            status = Status.UNBOUNDED
        elif result.solver.termination_condition == "infeasible":
            status = Status.INFEASIBLE
        else:
            status = Status.UNKNOWN

        upper_bound = result.problem.upper_bound
        iter_count = -1  # TODO: how to recover this with pyomo?
        solve_time = result.solver.wallclock_time
        x = np.array(
            [
                model.x[i].value if model.x[i].value is not None else 0.0
                for i in model.N
            ]
        )

        return Result(
            status,
            solve_time,
            iter_count,
            x,
            upper_bound,
            None,
        )

    def solve(
        self,
        datafit: MipDatafit,
        penalty: MipPenalty,
        A: NDArray,
        lmbd: float,
        x_init: Optional[NDArray] = None,
    ):
        """Solve an L0-regularized problem.

        Parameters
        ----------
        datafit: MipDatafit
            Problem datafit function.
        penalty: MipDatafit
            Problem penalty function.
        A: NDArray
            Problem matrix.
        lmbd: float
            Problem L0-norm weight parameter.
        x_init: NDArray, default=None
            Initial point for the solver.
        """

        optim = self.initialize_optimizer()
        model = self.build_model(datafit, penalty, A, lmbd)
        optim.set_instance(model)

        if x_init is not None:
            assert len(x_init) == A.shape[1]
            for i, xi in enumerate(x_init):
                model.x[i].set_value(xi)

        result = optim.solve(model, warmstart=True)

        return self.package_result(model, result)



"""Miscellaneous utilities."""

import numpy as np
from numpy.typing import NDArray

from el0ps.datafit import BaseDatafit
from el0ps.penalty import BasePenalty


def compute_lmbd_max(
    datafit: BaseDatafit, penalty: BasePenalty, A: NDArray
) -> float:
    r"""
    Return a value :math:`\lambda_{\max}` such that the all-zero vector is
    a solution of an L0-regularized problem whenever
    :math:`\lambda \geq \lambda_{\max}`.

    The problem is expressed as

    .. math::

        \textstyle\min_{\mathbf{x} \in \mathbb{R}^{n}} f(\mathbf{Ax}) + \lambda\|\mathbf{x}\|_0 + h(\mathbf{x})

    where :math:`f` is a :class:`el0ps.datafit.BaseDatafit` function,
    :math:`\mathbf{A} \in \mathbb{R}^{m \times n}` is a matrix, :math:`h` is a
    :class:`el0ps.penalty.BasePenalty` function, and :math:`\lambda` is a
    positive scalar.

    Parameters
    ----------
    datafit: BaseDatafit
        Problem datafit function.
    penalty: BasePenalty
        Problem penalty function.
    A: NDArray
        Problem matrix.

    Returns
    -------
    lmbd_max: float
        The value ``lmbd_max`` ensuring an all-zero solution.
    """  # noqa: E501
    w = np.zeros(A.shape[0])
    v = np.abs(A.T @ datafit.gradient(w))
    i = np.argmax(v)
    return penalty.conjugate(i, v[i])



"""Outer-Approximation solver for L0-regularized problems."""

import numpy as np
import pyomo.environ as pyo
import pyomo.kernel as pmo
import sys
from time import time
from typing import Optional, Union
from numba.experimental.jitclass.base import JitClassType
from numpy.typing import NDArray
from pyomo.opt import OptSolver

from el0ps.compilation import CompilableClass, compiled_clone
from el0ps.datafit import BaseDatafit
from el0ps.penalty import BasePenalty
from el0ps.solver import BaseSolver, Result, Status
from el0ps.solver.mip import _mip_optim_bindings


class OaSolver(BaseSolver):
    r"""Outer approximation solver for L0-regularized problems.

    The problem is expressed as

    .. math::

        \textstyle\min_{\mathbf{x} \in \mathbb{R}^{n}} f(\mathbf{Ax}) + \lambda\|\mathbf{x}\|_0 + h(\mathbf{x})

    where :math:`f` is a :class:`el0ps.datafit.BaseDatafit` function,
    :math:`\mathbf{A} \in \mathbb{R}^{m \times n}` is a matrix, :math:`h` is a
    :class:`el0ps.penalty.BasePenalty` function, and :math:`\lambda` is a
    positive scalar. This method is based on the work presented in "A unified
    approach to mixed-integer optimization problems with logical constraints"
    by D. Bertsimas et al. To use this solver, the optimizer specified in the
    ``optimizer_name`` parameter must be installed and accessible by
    `pyomo <https://pyomo.readthedocs.io/en/stable/>`_ which is the underlying
    library used to model the outer problem.

    Parameters
    ----------
    optimizer_name: str = "gurobi"
        Mixed-Integer Programming optimizer to use in the outer step. Available
        options are "cplex", "gurobi", and "mosek".
    relative_gap: float, default=1e-8
        Relative tolerance on the objective value.
    absolute_gap: float, default=0.0
        Absolute tolerance on the objective value.
    time_limit: float, default=None
        Limit in second on the solving time.
    iter_limit: int, default=None
        Limit on the number of outer-approximation steps performed.
    inner_iter_limit: int, default=None
        Maximum number of iterations for the inner steps optimization solver.
    inner_rel_tol: float, default=1e-8
        Relative tolerance on the objective value for the inner steps.
    verbose: bool, default=False
        Whether to toggle solver verbosity.
    keeptrace: bool, default=False
        Whether to store the solver trace.
    """  # noqa: E501

    _trace_keys = [
        "timer",
        "timer_outer",
        "timer_inner",
        "iter_count",
        "upper_bound",
        "lower_bound",
    ]

    def __init__(
        self,
        optimizer_name: str = "gurobi",
        relative_gap: float = 1e-8,
        absolute_gap: float = 0.0,
        time_limit: Optional[float] = None,
        iter_limit: Optional[int] = None,
        inner_iter_limit: Optional[int] = None,
        inner_rel_tol: float = 1e-8,
        verbose: bool = False,
        keeptrace: bool = False,
    ) -> None:
        self.optimizer_name = optimizer_name
        self.relative_gap = relative_gap
        self.absolute_gap = absolute_gap
        self.time_limit = time_limit if time_limit is not None else np.inf
        self.iter_limit = iter_limit if iter_limit is not None else sys.maxsize
        self.inner_iter_limit = (
            inner_iter_limit if inner_iter_limit is not None else sys.maxsize
        )
        self.inner_rel_tol = inner_rel_tol
        self.verbose = verbose
        self.keeptrace = keeptrace

    def __str__(self):
        return "OaSolver"

    def initialize_optimizer(self) -> OptSolver:
        if self.optimizer_name in _mip_optim_bindings:
            bindings = _mip_optim_bindings[self.optimizer_name]
            optim = pyo.SolverFactory(bindings["optimizer_name"])
            optim.options[bindings["time_limit"]] = self.time_limit
            optim.options[bindings["relative_gap"]] = self.relative_gap
            optim.options[bindings["absolute_gap"]] = self.absolute_gap
            optim.options[bindings["verbose"]] = int(self.verbose)
        else:
            raise ValueError(
                "Solver {} not supported. Available ones are: {}".format(
                    self.optimizer_name,
                    _mip_optim_bindings.keys(),
                )
            )
        return optim

    def initialize_inner_solver(self):
        self.lipschitz = self.datafit.gradient_lipschitz_constant()
        self.A_colnorm = np.linalg.norm(self.A, ord=2, axis=0)
        self.stepsize = 1.0 / (self.lipschitz * self.A_colnorm**2)

    def build_model(self) -> None:
        model = pmo.block()
        model.N = range(self.A.shape[1])
        model.z = pmo.variable_dict()
        model.t = pmo.variable(domain=pmo.Reals)
        for i in model.N:
            model.z[i] = pmo.variable(domain=pmo.Binary)
        model.obj = pmo.objective(
            self.lmbd * sum(model.z[i] for i in model.N) + model.t
        )
        model.t_con = pmo.constraint_dict()
        return model

    def add_cut(self, optim, model, z: NDArray, f: float, g: NDArray) -> None:
        self.cuts_count += 1
        model.t_con[self.cuts_count] = pmo.constraint(
            model.t >= f + sum(g[i] * (model.z[i] - z[i]) for i in model.N)
        )
        optim.add_constraint(model.t_con[self.cuts_count])

    def outer_solve(self, optim, model):
        start_time_outer = time()
        result = optim.solve(model, warmstart=True, save_results=False)
        self.timer_outer += time() - start_time_outer
        zk = np.array([model.z[i].value for i in model.N])
        vk = result.problem.lower_bound
        return zk, vk

    def inner_solve(self, z):

        start_time_inner = time()

        Z = z != 0.0
        x = np.zeros(self.A.shape[1])
        x[Z] = np.copy(self.x[Z])
        w = self.A[:, Z] @ x[Z]
        u = -self.datafit.gradient(w)

        pv = np.inf
        for _ in range(self.inner_iter_limit):
            pv_old = pv
            for i in np.flatnonzero(Z):
                ai = self.A[:, i]
                xi = x[i]
                ci = xi + self.stepsize[i] * np.dot(ai, u)
                x[i] = self.penalty.prox(i, ci, self.stepsize[i])
                if x[i] != xi:
                    w += (x[i] - xi) * ai
                    u[:] = -self.datafit.gradient(w)
            pv = self.datafit.value(w) + sum(
                self.penalty.value(xi) for xi in x
            )
            rtol = np.abs(pv - pv_old) / (np.abs(pv) + 1e-10)
            if rtol < self.inner_rel_tol:
                break

        v = self.A.T @ u
        g = np.array([self.penalty.conjugate(i, vi) for i, vi in enumerate(v)])
        dv = -self.datafit.conjugate(-u) - np.dot(z, g)
        self.timer_inner += time() - start_time_inner

        return x, dv, -g

    @property
    def abs_gap(self):
        return self.upper_bound - self.lower_bound

    @property
    def rel_gap(self):
        return (self.upper_bound - self.lower_bound) / (
            np.abs(self.upper_bound) + 1e-10
        )

    @property
    def timer(self):
        return time() - self.start_time

    def print_header(self):
        s = "-" * 50 + "\n"
        s += "|"
        s += " {:>6}".format("Iters")
        s += " {:>6}".format("Timer")
        s += " {:>6}".format("Lower")
        s += " {:>6}".format("Upper")
        s += " {:>9}".format("Abs gap")
        s += " {:>9}".format("Rel gap")
        s += "|" + "\n"
        s += "-" * 50
        print(s)

    def print_progress(self):
        s = "|"
        s += " {:>6d}".format(self.iter_count)
        s += " {:>6.2f}".format(self.timer)
        s += " {:>6.2f}".format(self.lower_bound)
        s += " {:>6.2f}".format(self.upper_bound)
        s += " {:>9.2e}".format(self.abs_gap)
        s += " {:>9.2e}".format(self.rel_gap)
        s += "|"
        print(s)

    def print_footer(self):
        s = "-" * 50
        print(s)

    def update_bounds(self, zk, vk, xk, fk) -> None:
        if self.lower_bound < vk:
            self.lower_bound = vk
        if self.upper_bound > fk + self.lmbd * np.sum(zk):
            self.upper_bound = fk + self.lmbd * np.sum(zk)
            self.x = xk

    def update_trace(self) -> None:
        for key in self._trace_keys:
            self.trace[key].append(getattr(self, key))

    def can_continue(self):
        if self.timer >= self.time_limit:
            self.status = Status.TIME_LIMIT
        elif self.iter_count >= self.iter_limit:
            self.status = Status.ITER_LIMIT
        elif self.rel_gap < self.relative_gap:
            self.status = Status.OPTIMAL
        return self.status == Status.RUNNING

    def solve(
        self,
        datafit: Union[BaseDatafit, CompilableClass, JitClassType],
        penalty: Union[BasePenalty, CompilableClass, JitClassType],
        A: NDArray,
        lmbd: float,
        x_init: Optional[NDArray] = None,
    ):
        """Solve an L0-regularized problem.

        Parameters
        ----------
        datafit: Union[BaseDatafit, JitClassType]
            Problem datafit function. If not already JIT-compiled the solver
            automatically compiles if it derives from the
            :class:`el0ps.compilation.CompilableClass`.
        penalty: Union[BaseDatafit, JitClassType]
            Problem penalty function. If not already JIT-compiled the solver
            automatically compiles if it derives from the
            :class:`el0ps.compilation.CompilableClass`.
        A: NDArray
            Problem matrix.
        lmbd: float
            Problem L0-norm weight parameter.
        x_init: NDArray, default=None
            Initial point for the solver. If `None`, the solver initializes it
            to the all-zero vector.
        """

        if isinstance(datafit, CompilableClass):
            if not isinstance(datafit, JitClassType):
                datafit = compiled_clone(datafit)
        if isinstance(penalty, CompilableClass):
            if not isinstance(penalty, JitClassType):
                penalty = compiled_clone(penalty)

        self.datafit = datafit
        self.penalty = penalty
        self.A = A
        self.lmbd = lmbd

        self.status = Status.RUNNING
        self.iter_count = 0
        self.cuts_count = 0
        self.timer_outer = 0.0
        self.timer_inner = 0.0
        self.x = x_init if x_init is not None else np.zeros(self.A.shape[1])
        self.upper_bound = (
            self.datafit.value(self.A @ self.x)
            + self.lmbd * np.linalg.norm(self.x, 0)
            + sum(self.penalty.value(xi) for xi in self.x)
        )
        self.lower_bound = -np.inf
        self.trace = {k: [] for k in self._trace_keys}

        self.start_time = time()

        optim = self.initialize_optimizer()
        model = self.build_model()
        optim.set_instance(model)

        self.initialize_inner_solver()

        if self.verbose:
            self.print_header()

        z0 = np.array(self.x != 0.0, dtype=int)
        _, f0, g0 = self.inner_solve(z0)
        self.add_cut(optim, model, z0, f0, g0)

        while self.can_continue():
            self.iter_count += 1
            zk, vk = self.outer_solve(optim, model)
            xk, fk, gk = self.inner_solve(zk)
            self.add_cut(optim, model, zk, fk, gk)
            self.update_bounds(zk, vk, xk, fk)
            if self.verbose:
                self.print_progress()
            if self.keeptrace:
                self.update_trace()

        if self.verbose:
            self.print_footer()

        return Result(
            self.status,
            self.timer,
            self.iter_count,
            self.x,
            self.upper_bound,
            self.trace,
        )



import numpy as np
import pyomo.kernel as pmo
from numpy.typing import NDArray
from numba import float64

from el0ps.compilation import CompilableClass
from el0ps.penalty.base import BasePenalty, MipPenalty


class BigmPositiveL2norm(CompilableClass, BasePenalty, MipPenalty):
    r"""Positive big-M plus L2-norm :class:`BasePenalty` penalty function.

    The splitting terms are expressed as

    .. math::
        h_i(x_i) = \begin{cases}
        \beta x_i^2 & \text{if } 0 \leq x_i \leq M \\
        +\infty & \text{otherwise}
        \end{cases}

    for some :math:`M > 0` and :math:`\beta > 0`.

    Parameters
    ----------
    M: float
        Big-M value.
    beta: float
        L2-norm weight.
    """

    def __init__(self, M: float, beta: float) -> None:
        self.M = M
        self.beta = beta

    def __str__(self) -> str:
        return "BigmPositiveL2norm"

    def get_spec(self) -> tuple:
        spec = (
            ("M", float64),
            ("beta", float64),
        )
        return spec

    def params_to_dict(self) -> dict:
        return dict(M=self.M, beta=self.beta)

    def value(self, i: int, x: float) -> float:
        return self.beta * x**2 if (x >= 0.0) and (x <= self.M) else np.inf

    def conjugate(self, i: int, x: float) -> float:
        r = np.maximum(np.minimum(x / (2.0 * self.beta), self.M), 0.0)
        return x * r - self.beta * r**2

    def prox(self, i: int, x: float, eta: float) -> float:
        v = x / (1.0 + 2.0 * eta * self.beta)
        return np.maximum(np.minimum(v, self.M), 0.0)

    def subdiff(self, i: int, x: float) -> NDArray:
        if x == 0.0:
            return [-np.inf, 0.0]
        elif 0.0 < x < self.M:
            s = 2.0 * self.beta * x
            return [s, s]
        elif x == self.M:
            return [2.0 * self.beta * x, np.inf]
        else:
            return [np.nan, np.nan]

    def conjugate_subdiff(self, i: int, x: float) -> NDArray:
        s = np.maximum(np.minimum(x / (2.0 * self.beta), self.M), 0.0)
        return [s, s]

    def param_slope_pos(self, i: int, lmbd: float) -> float:
        if lmbd < self.beta * self.M**2:
            return np.sqrt(4.0 * lmbd * self.beta)
        else:
            return (lmbd / self.M) + self.beta * self.M

    def param_slope_neg(self, i: int, lmbd: float) -> float:
        return -np.inf

    def param_limit_pos(self, i: int, lmbd: float) -> float:
        if lmbd < self.beta * self.M**2:
            return np.sqrt(lmbd / self.beta)
        else:
            return self.M

    def param_limit_neg(self, i: int, lmbd: float) -> float:
        return 0.0

    def param_bndry_pos(self, i, lmbd):
        if lmbd < self.beta * self.M**2:
            return np.sqrt(4.0 * lmbd * self.beta)
        else:
            return np.inf

    def param_bndry_neg(self, i, lmbd):
        return -np.inf

    def bind_model(self, model: pmo.block) -> None:

        model.h1_var = pmo.variable_dict()
        for i in model.N:
            model.h1_var[i] = pmo.variable(
                domain=pmo.NonNegativeReals, ub=self.M**2
            )

        model.hpos_con = pmo.constraint_dict()
        model.hneg_con = pmo.constraint_dict()
        model.h1_con = pmo.constraint_dict()
        for i in model.N:
            model.hpos_con[i] = pmo.constraint(
                model.x[i] <= self.M * model.z[i]
            )
            model.hneg_con[i] = pmo.constraint(model.x[i] >= 0.0)
            model.h1_con[i] = pmo.conic.rotated_quadratic(
                model.h1_var[i], model.z[i], [model.x[i]]
            )
        model.h_con = pmo.constraint(
            model.h >= 2.0 * self.beta * sum(model.h1_var[i] for i in model.N)
        )



"""L0-penalized problem solvers."""

from .base import BaseSolver, Result, Status
from .bnb import BoundSolver, BnbSolver
from .mip import MipSolver
from .oa import OaSolver

__all__ = [
    "BaseSolver",
    "Result",
    "Status",
    "BoundSolver",
    "BnbSolver",
    "MipSolver",
    "OaSolver",
]



import numpy as np
from numpy.typing import NDArray
from numba import float64

from el0ps.compilation import CompilableClass

from .base import BasePenalty


class PositiveL1norm(CompilableClass, BasePenalty):
    r"""Positive L1-norm :class:`BasePenalty` penalty function.

    The splitting terms are expressed as

    .. math::
        h_i(x_i) = \begin{cases}
        \alpha x_i & \text{if } x_i \geq 0 \\
        +\infty & \text{otherwise}
        \end{cases}

    for some :math:`\alpha > 0`.

    Parameters
    ----------
    alpha: float
        L1-norm weight.
    """

    def __init__(self, alpha: float) -> None:
        self.alpha = alpha

    def __str__(self) -> str:
        return "PositiveL1norm"

    def get_spec(self) -> tuple:
        spec = (("alpha", float64),)
        return spec

    def params_to_dict(self) -> dict:
        return dict(alpha=self.alpha)

    def value(self, i: int, x: float) -> float:
        return self.alpha * x if x >= 0.0 else np.inf

    def conjugate(self, i: int, x: float) -> float:
        return 0.0 if x <= self.alpha else np.inf

    def prox(self, i: int, x: float, eta: float) -> float:
        return np.maximum(0.0, x - eta * self.alpha)

    def subdiff(self, i: int, x: float) -> NDArray:
        if x == 0:
            return [-np.inf, self.alpha]
        elif x > 0:
            s = self.alpha
            return [s, s]
        else:
            return [np.nan, np.nan]

    def conjugate_subdiff(self, i: int, x: float) -> NDArray:
        if x < self.alpha:
            return [0.0, 0.0]
        elif x == self.alpha:
            return [0.0, np.inf]
        else:
            return [np.nan, np.nan]

    def param_slope_pos(self, i: int, lmbd: float) -> float:
        return self.alpha

    def param_slope_neg(self, i: int, lmbd: float) -> float:
        return -np.inf

    def param_limit_pos(self, i: int, lmbd: float) -> float:
        return np.inf

    def param_limit_neg(self, i: int, lmbd: float) -> float:
        return 0.0

    def param_bndry_pos(self, i, lmbd):
        return self.alpha

    def param_bndry_neg(self, i, lmbd):
        return -np.inf



import numpy as np
import pyomo.kernel as pmo
from numpy.typing import NDArray
from numba import float64

from el0ps.compilation import CompilableClass

from .base import BasePenalty, MipPenalty


class PositiveL2norm(CompilableClass, BasePenalty, MipPenalty):
    r"""Positive L2-norm :class:`BasePenalty` penalty function.

    The splitting terms are expressed as

    .. math::
        h_i(x_i) = \begin{cases}
        \beta x_i^2 & \text{if } x_i \geq 0 \\
        +\infty & \text{otherwise}
        \end{cases}

    for some :math:`\beta > 0`.

    Parameters
    ----------
    beta: float
        L2-norm weight.
    """

    def __init__(self, beta: float) -> None:
        self.beta = beta

    def __str__(self) -> str:
        return "PositiveL2norm"

    def get_spec(self) -> tuple:
        spec = (("beta", float64),)
        return spec

    def params_to_dict(self) -> dict:
        return dict(beta=self.beta)

    def value(self, i: int, x: float) -> float:
        return self.beta * x**2 if x >= 0.0 else np.inf

    def conjugate(self, i: int, x: float) -> float:
        return np.maximum(x, 0.0) ** 2 / (4.0 * self.beta)

    def prox(self, i: int, x: float, eta: float) -> float:
        return np.maximum(x, 0.0) / (1.0 + 2.0 * self.beta * eta)

    def subdiff(self, i: int, x: float) -> NDArray:
        if x == 0.0:
            return [-np.inf, 0.0]
        elif x > 0.0:
            s = 2.0 * self.beta * x
            return [s, s]
        else:
            return [np.nan, np.nan]

    def conjugate_subdiff(self, i: int, x: float) -> NDArray:
        if x > 0.0:
            s = x / (2.0 * self.beta)
            return [s, s]
        else:
            s = 0.0
            return [s, s]

    def param_slope_pos(self, i: int, lmbd: float) -> float:
        return 2.0 * np.sqrt(lmbd * self.beta)

    def param_slope_neg(self, i: int, lmbd: float) -> float:
        return -np.inf

    def param_limit_pos(self, i: int, lmbd: float) -> float:
        return np.sqrt(lmbd / self.beta)

    def param_limit_neg(self, i: int, lmbd: float) -> float:
        return 0.0

    def param_bndry_pos(self, i, lmbd):
        return 2.0 * np.sqrt(lmbd * self.beta)

    def param_bndry_neg(self, i, lmbd):
        return -np.inf

    def bind_model(self, model: pmo.block) -> None:
        model.h1_var = pmo.variable_dict()
        for i in model.N:
            model.h1_var[i] = pmo.variable(domain=pmo.NonNegativeReals)

        model.h1_con = pmo.constraint_dict()
        model.hpos_con = pmo.constraint_dict()
        for i in model.N:
            model.h1_con[i] = pmo.conic.rotated_quadratic(
                model.h1_var[i], model.z[i], [model.x[i]]
            )
            model.hpos_con[i] = pmo.constraint(model.x[i] >= 0.0)
        model.h_con = pmo.constraint(
            model.h >= 2.0 * self.beta * sum(model.h1_var[i] for i in model.N)
        )



"""Penalty functions."""

from .base import (
    BasePenalty,
    MipPenalty,
    SymmetricPenalty,
    compute_param_slope_pos,
    compute_param_slope_neg,
)
from .bigm import Bigm
from .bigml1norm import BigmL1norm
from .bigml2norm import BigmL2norm
from .bigmpositivel1norm import BigmPositiveL1norm
from .bigmpositivel2norm import BigmPositiveL2norm
from .bounds import Bounds
from .l1norm import L1norm
from .l2norm import L2norm
from .l1l2norm import L1L2norm
from .positivel1norm import PositiveL1norm
from .positivel2norm import PositiveL2norm


__all__ = [
    "BasePenalty",
    "MipPenalty",
    "SymmetricPenalty",
    "compute_param_slope_pos",
    "compute_param_slope_neg",
    "Bigm",
    "BigmL1norm",
    "BigmL2norm",
    "BigmPositiveL1norm",
    "BigmPositiveL2norm",
    "Bounds",
    "L1norm",
    "L2norm",
    "L1L2norm",
    "PositiveL1norm",
    "PositiveL2norm",
]



import numpy as np
import pyomo.kernel as pmo
from numpy.typing import NDArray
from numba import float64

from el0ps.compilation import CompilableClass
from el0ps.penalty.base import BasePenalty, MipPenalty


class Bounds(CompilableClass, BasePenalty, MipPenalty):
    r"""Bound-constraint :class:`BasePenalty` penalty function.

    The splitting terms are expressed as

    .. math::
        h_i(x_i) = \begin{cases}
        0 & \text{if } x^{\text{lb}}_i \leq x_i \leq x^{\text{ub}}_i \\
        +\infty & \text{otherwise}
        \end{cases}

    for some :math:`x^{\text{lb}}_i < 0` and :math:`x^{\text{ub}}_i > 0`.

    Parameters
    ----------
    x_lb: NDArray
        Vector or lower bounds.
    x_ub: NDArray
        Vector of upper bounds.
    """

    def __init__(self, x_lb: NDArray, x_ub: NDArray) -> None:
        self.x_lb = x_lb
        self.x_ub = x_ub

    def __str__(self) -> str:
        return "Bounds"

    def get_spec(self) -> tuple:
        spec = (("x_lb", float64[:]), ("x_ub", float64[:]))
        return spec

    def params_to_dict(self) -> dict:
        return dict(x_lb=self.x_lb, x_ub=self.x_ub)

    def value(self, i: int, x: float) -> float:
        return 0.0 if (self.x_lb[i] <= x <= self.x_ub[i]) else np.inf

    def conjugate(self, i: int, x: float) -> float:
        if x >= 0.0:
            return self.x_ub[i] * x
        else:
            return self.x_lb[i] * x

    def prox(self, i: int, x: float, eta: float) -> float:
        return np.maximum(np.minimum(x, self.x_ub[i]), self.x_lb[i])

    def subdiff(self, i: int, x: float) -> NDArray:
        if self.x_lb[i] < x < self.x_ub[i]:
            return [0.0, 0.0]
        elif x == self.x_lb[i]:
            return [-np.inf, 0.0]
        elif x == self.x_ub[i]:
            return [0.0, np.inf]
        else:
            return [np.nan, np.nan]

    def conjugate_subdiff(self, i: int, x: float) -> NDArray:
        if x == 0.0:
            return [self.x_lb[i], self.x_ub[i]]
        elif x > 0.0:
            s = self.x_ub[i]
            return [s, s]
        else:
            s = self.x_lb[i]
            return [s, s]

    def param_slope_pos(self, i: int, lmbd: float) -> float:
        if self.x_ub[i] == 0.0:
            return np.inf
        return lmbd / self.x_ub[i]

    def param_slope_neg(self, i: int, lmbd: float) -> float:
        if self.x_lb[i] == 0.0:
            return -np.inf
        return lmbd / self.x_lb[i]

    def param_limit_pos(self, i: int, lmbd: float) -> float:
        return self.x_ub[i]

    def param_limit_neg(self, i: int, lmbd: float) -> float:
        return self.x_lb[i]

    def param_bndry_pos(self, i, lmbd):
        return np.inf

    def param_bndry_neg(self, i, lmbd):
        return -np.inf

    def bind_model(self, model: pmo.block) -> None:
        model.hpos_con = pmo.constraint_dict()
        model.hneg_con = pmo.constraint_dict()
        for i in model.N:
            model.hpos_con[i] = pmo.constraint(
                model.x[i] <= self.x_ub[i] * model.z[i]
            )
            model.hneg_con[i] = pmo.constraint(
                model.x[i] >= self.x_lb[i] * model.z[i]
            )
        model.h_con = pmo.constraint(model.h >= 0.0)



import numpy as np
import pyomo.kernel as pmo
from numba import float64
from numpy.typing import NDArray

from el0ps.compilation import CompilableClass
from el0ps.datafit.base import BaseDatafit, MipDatafit


class Leastsquares(CompilableClass, BaseDatafit, MipDatafit):
    r"""Least-squares datafit function.

    The function is defined as

    .. math::

        f(\mathbf{w}) = \sum_{i=1}^m \tfrac{1}{2}(y_i - w_i)^2

    for some :math:`\mathbf{y} \in \mathbb{R}^m`.

    Parameters
    ----------
    y : NDArray
        Data vector.
    """

    def __init__(self, y: NDArray) -> None:
        self.y = y
        self.L = 1.0

    def __str__(self) -> str:
        return "Leastsquares"

    def get_spec(self) -> tuple:
        spec = (("y", float64[::1]), ("L", float64))
        return spec

    def params_to_dict(self) -> dict:
        return dict(y=self.y)

    def value(self, w: NDArray) -> float:
        v = w - self.y
        return 0.5 * np.dot(v, v)

    def conjugate(self, w: NDArray) -> float:
        return 0.5 * np.dot(w, w) + np.dot(w, self.y)

    def gradient(self, w: NDArray) -> NDArray:
        return w - self.y

    def gradient_lipschitz_constant(self) -> float:
        return self.L

    def bind_model(self, model: pmo.block) -> None:
        model.c_var = pmo.variable()
        model.c_con = pmo.constraint(model.c_var == 1.0)
        model.r_var = pmo.variable_dict()
        model.r_con = pmo.constraint_dict()
        for j in model.M:
            model.r_var[j] = pmo.variable(domain=pmo.Reals)
            model.r_con[j] = pmo.constraint(
                model.r_var[j] == model.w[j] - self.y[j]
            )
        model.f_con = pmo.conic.rotated_quadratic(
            model.f,
            model.c_var,
            [model.r_var[j] for j in model.M],
        )



import numpy as np
import pyomo.kernel as pmo
from numpy.typing import NDArray
from numba import float64

from el0ps.compilation import CompilableClass
from el0ps.penalty.base import SymmetricPenalty, MipPenalty


class BigmL1norm(CompilableClass, SymmetricPenalty, MipPenalty):
    r"""Big-M plus L1-norm :class:`BasePenalty` penalty function.

    The splitting terms are expressed as

    .. math::
        h_i(x_i) = \begin{cases}
        \alpha|x_i| & \text{if } |x_i| \leq M \\
        +\infty & \text{otherwise}
        \end{cases}

    for some :math:`M > 0` and :math:`\alpha > 0`.

    Parameters
    ----------
    M: float
        Big-M value.
    alpha: float
        L1-norm weight.
    """

    def __init__(self, M: float, alpha: float) -> None:
        self.M = M
        self.alpha = alpha

    def __str__(self) -> str:
        return "BigmL1norm"

    def get_spec(self) -> tuple:
        spec = (
            ("M", float64),
            ("alpha", float64),
        )
        return spec

    def params_to_dict(self) -> dict:
        return dict(M=self.M, alpha=self.alpha)

    def value(self, i: int, x: float) -> float:
        xabs = np.abs(x)
        return self.alpha * xabs if xabs <= self.M else np.inf

    def conjugate(self, i: int, x: float) -> float:
        return self.M * np.maximum(np.abs(x) - self.alpha, 0.0)

    def prox(self, i: int, x: float, eta: float) -> float:
        v = np.abs(x) - eta * self.alpha
        return np.sign(x) * np.maximum(np.minimum(v, self.M), 0.0)

    def subdiff(self, i: int, x: float) -> NDArray:
        if x == 0.0:
            return [-self.alpha, self.alpha]
        elif np.abs(x) < self.M:
            return [self.alpha, self.alpha]
        elif x == -self.M:
            return [-np.inf, -self.alpha]
        elif x == self.M:
            return [self.alpha, np.inf]
        else:
            return [np.nan, np.nan]

    def conjugate_subdiff(self, i: int, x: float) -> NDArray:
        if np.abs(x) < self.alpha:
            return [0.0, 0.0]
        elif x == self.alpha:
            return [0.0, self.M]
        elif x == -self.alpha:
            return [-self.M, 0.0]
        else:
            s = np.sign(x) * self.M
            return [s, s]

    def param_slope(self, i: int, lmbd: float) -> float:
        return (lmbd / self.M) + self.alpha

    def param_limit(self, i: int, lmbd: float) -> float:
        return self.M

    def param_bndry(self, i, lmbd):
        return np.inf

    def bind_model(self, model: pmo.block) -> None:

        model.h1_var = pmo.variable_dict()
        for i in model.N:
            model.h1_var[i] = pmo.variable(domain=pmo.NonNegativeReals)

        model.hpos_con = pmo.constraint_dict()
        model.hneg_con = pmo.constraint_dict()
        model.h1pos_con = pmo.constraint_dict()
        model.h1neg_con = pmo.constraint_dict()
        for i in model.N:
            model.hpos_con[i] = pmo.constraint(
                model.x[i] <= self.M * model.z[i]
            )
            model.hneg_con[i] = pmo.constraint(
                model.x[i] >= -self.M * model.z[i]
            )
            model.h1pos_con[i] = pmo.constraint(model.h1_var[i] >= model.x[i])
            model.h1neg_con[i] = pmo.constraint(model.h1_var[i] >= -model.x[i])
        model.h_con = pmo.constraint(
            model.h >= self.alpha * sum(model.h1_var[i] for i in model.N)
        )



import numpy as np
import pyomo.kernel as pmo
from numba import float64
from numpy.typing import NDArray

from el0ps.compilation import CompilableClass
from el0ps.datafit.base import BaseDatafit, MipDatafit


class Logistic(CompilableClass, BaseDatafit, MipDatafit):
    r"""Logistic datafit function.

    The function is defined as

    .. math::

        f(\mathbf{w}) = \sum_{i=1}^m \log(1 + \exp(-y_i w_i))

    for some :math:`\mathbf{y} \in \mathbb{R}^m`.

    Parameters
    ----------
    y : NDArray
        Data vector.
    """

    def __init__(self, y: NDArray) -> None:
        self.y = y
        self.L = 0.25

    def __str__(self) -> str:
        return "Logistic"

    def get_spec(self) -> tuple:
        spec = (("y", float64[::1]), ("L", float64))
        return spec

    def params_to_dict(self) -> dict:
        return dict(y=self.y)

    def value(self, w: NDArray) -> float:
        return np.sum(np.log(1.0 + np.exp(-self.y * w)))

    def conjugate(self, w: NDArray) -> float:
        c = -(w * self.y)
        if not np.all((0.0 < c) & (c < 1.0)):
            return np.inf
        r = 1.0 - c
        return np.dot(c, np.log(c)) + np.dot(r, np.log(r))

    def gradient(self, w: NDArray) -> NDArray:
        return -self.y / (1.0 + np.exp(self.y * w))

    def gradient_lipschitz_constant(self) -> float:
        return self.L

    def bind_model(self, model: pmo.block) -> None:
        model.c1_var = pmo.variable_dict()
        model.c2_var = pmo.variable_dict()
        model.c3_var = pmo.variable_dict()
        model.f1_var = pmo.variable_dict()
        model.f2_var = pmo.variable_dict()
        model.f3_var = pmo.variable_dict()
        for j in model.M:
            model.c1_var[j] = pmo.variable(domain=pmo.Reals)
            model.c2_var[j] = pmo.variable(domain=pmo.Reals)
            model.c3_var[j] = pmo.variable(domain=pmo.Reals)
            model.f1_var[j] = pmo.variable(domain=pmo.Reals)
            model.f2_var[j] = pmo.variable(domain=pmo.Reals)
            model.f3_var[j] = pmo.variable(domain=pmo.Reals)

        model.c1_con = pmo.constraint_dict()
        model.c2_con = pmo.constraint_dict()
        model.c3_con = pmo.constraint_dict()
        model.f1f2_con = pmo.constraint_dict()
        model.f1f3_con = pmo.constraint_dict()
        model.f2f3_con = pmo.constraint_dict()
        for j in model.M:
            model.c1_con[j] = pmo.constraint(model.c1_var[j] == 1.0)
            model.c2_con[j] = pmo.constraint(
                model.c2_var[j] == self.y[j] * model.w[j] - model.f1_var[j]
            )
            model.c3_con[j] = pmo.constraint(
                model.c3_var[j] == -model.f1_var[j]
            )
            model.f1f2_con[j] = pmo.conic.primal_exponential(
                model.f2_var[j],
                model.c1_var[j],
                model.c2_var[j],
            )
            model.f1f3_con[j] = pmo.conic.primal_exponential(
                model.f3_var[j],
                model.c1_var[j],
                model.c3_var[j],
            )
            model.f2f3_con[j] = pmo.constraint(
                model.f2_var[j] + model.f3_var[j] <= 1.0
            )
        model.f_con = pmo.constraint(
            model.f >= sum(model.f1_var[j] for j in model.M)
        )
